import torch
import numpy as np

from .galaxy_model_object import Galaxy_Model
from .psf_model_object import PSF_Model
from ._shared_methods import parametric_initialize, select_target
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.parametric_profiles import moffat_np
from ..utils.conversions.functions import moffat_I0_to_flux, general_uncertainty_prop

__all__ = ["Moffat_Galaxy", "Moffat_PSF"]


def _x0_func(model_params, R, F):
    return 2.0, R[4], F[0]


def _wrap_moffat(R, n, rd, i0):
    return moffat_np(R, n, rd, 10 ** (i0))


class Moffat_Galaxy(Galaxy_Model):
    """basic galaxy model with a Moffat profile for the radial light
    profile. The functional form of the Moffat profile is defined as:

    I(R) = I0 / (1 + (R/Rd)^2)^n

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, I0 is the central flux
    density, Rd is the scale length for the profile, and n is the
    concentration index which controls the shape of the profile.

    Parameters:
        n: Concentration index which controls the shape of the brightness profile
        I0: brightness at the center of the profile, represented as the log of the brightness divided by pixel scale squared.
        Rd: scale length radius

    """

    model_type = f"moffat {Galaxy_Model.model_type}"
    parameter_specs = {
        "n": {"units": "none", "limits": (0.1, 10), "uncertainty": 0.05},
        "Rd": {"units": "arcsec", "limits": (0, None)},
        "I0": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("n", "Rd", "I0")
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self, parameters, target, _wrap_moffat, ("n", "Rd", "I0"), _x0_func
        )

    @default_internal
    def total_flux(self, parameters=None):
        return moffat_I0_to_flux(
            10 ** parameters["I0"].value,
            parameters["n"].value,
            parameters["Rd"].value,
            parameters["q"].value,
        )
    @default_internal
    def total_flux_uncertainty(self, parameters=None):
        return general_uncertainty_prop(
            (10 ** parameters["I0"].value,
             parameters["n"].value,
             parameters["Rd"].value,
             parameters["q"].value
            ),
            ((10 ** parameters["I0"].value) * parameters["I0"].uncertainty * torch.log(10*torch.ones_like(parameters["I0"].value)),
             parameters["n"].uncertainty,
             parameters["Rd"].uncertainty,
             parameters["q"].uncertainty
            ),
            moffat_I0_to_flux
        )
    
    from ._shared_methods import moffat_radial_model as radial_model


class Moffat_PSF(PSF_Model):
    """basic point source model with a Moffat profile for the radial light
    profile. The functional form of the Moffat profile is defined as:

    I(R) = I0 / (1 + (R/Rd)^2)^n

    where I(R) is the brightness profile as a function of semi-major
    axis, R is the semi-major axis length, I0 is the central flux
    density, Rd is the scale length for the profile, and n is the
    concentration index which controls the shape of the profile.

    Parameters:
        n: Concentration index which controls the shape of the brightness profile
        I0: brightness at the center of the profile, represented as the log of the brightness divided by pixel scale squared.
        Rd: scale length radius

    """

    model_type = f"moffat {PSF_Model.model_type}"
    parameter_specs = {
        "n": {"units": "none", "limits": (0.1, 10), "uncertainty": 0.05},
        "Rd": {"units": "arcsec", "limits": (0, None)},
        "I0": {"units": "log10(flux/arcsec^2)", "value": 0., "locked": True},
    }
    _parameter_order = PSF_Model._parameter_order + ("n", "Rd", "I0")
    useable = True
    model_integrated = False
    
    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        parametric_initialize(
            self, parameters, target, _wrap_moffat, ("n", "Rd", "I0"), _x0_func
        )

    from ._shared_methods import moffat_radial_model as radial_model

    @default_internal
    def total_flux(self, parameters=None):
        return moffat_I0_to_flux(
            10 ** parameters["I0"].value,
            parameters["n"].value,
            parameters["Rd"].value,
            torch.ones_like(parameters["n"].value),
        )
    @default_internal
    def total_flux_uncertainty(self, parameters=None):
        return general_uncertainty_prop(
            (10 ** parameters["I0"].value,
             parameters["n"].value,
             parameters["Rd"].value,
             torch.ones_like(parameters["n"].value)
            ),
            ((10 ** parameters["I0"].value) * parameters["I0"].uncertainty * torch.log(10*torch.ones_like(parameters["I0"].value)),
             parameters["n"].uncertainty,
             parameters["Rd"].uncertainty,
             torch.zeros_like(parameters["n"].value)
            ),
            moffat_I0_to_flux
        )
    
    from ._shared_methods import radial_evaluate_model as evaluate_model



from .base import AstroPhotError

__all__ = ("InvalidWindow", "ConflicingWCS", "InvalidData", "InvalidImage", "InvalidWCS")

class InvalidWindow(AstroPhotError):
    """
    Raised whenever a window is misspecified
    """
    ...

class ConflicingWCS(InvalidWindow):
    """
    Raised when windows are compared and have WCS prescriptions which do not agree
    """
    ...
    
class InvalidData(AstroPhotError):
    """
    Raised when an image object can't determine the data it is holding.
    """
    ...

class InvalidImage(AstroPhotError):
    """
    Raised when an image object cannot be used as given.
    """
    ...

class InvalidWCS(AstroPhotError):
    """
    Raised when the WCS is not appropriate as given.
    """
    ...




import torch
import numpy as np
import astrophot as ap
from astropy.wcs import WCS

def get_astropy_wcs():
    hdr = {
        "SIMPLE": "T",
        "NAXIS": 2,
        "NAXIS1": 180,
        "NAXIS2": 180,
        "CTYPE1": "RA---TAN",
        "CTYPE2": "DEC--TAN",
        "CRVAL1": 195.0588,
        "CRVAL2": 28.0608,
        "CRPIX1": 90.5,
        "CRPIX2": 90.5,
        "CD1_1": -0.000416666666666667,
        "CD1_2": 0.,
        "CD2_1": 0.,
        "CD2_2": 0.000416666666666667,
        "IMAGEW": 180.,
        "IMAGEH": 180.,
    }
    return WCS(hdr)


def make_basic_sersic(
    N=50,
    M=50,
    pixelscale=0.8,
    x=24.5,
    y=25.4,
    PA=45 * np.pi / 180,
    q=0.6,
    n=2,
    Re=7.1,
    Ie=0,
    rand=12345,
):

    np.random.seed(rand)
    mask = np.zeros((N, M), dtype = bool)
    mask[0][0] = True
    target = ap.image.Target_Image(
        data=np.zeros((N, M)),
        pixelscale=pixelscale,
        psf=ap.utils.initialize.gaussian_psf(2 / pixelscale, 11, pixelscale),
        mask = mask,
    )

    MODEL = ap.models.Sersic_Galaxy(
        name="basic sersic model",
        target=target,
        parameters={
            "center": [x, y],
            "PA": PA,
            "q": q,
            "n": n,
            "Re": Re,
            "Ie": Ie,
        },
    )

    img = MODEL().data.detach().cpu().numpy()
    target.data = (
        img
        + np.random.normal(scale=0.1, size=img.shape)
        + np.random.normal(scale=np.sqrt(img) / 10)
    )
    target.variance = 0.1 ** 2 + img / 100

    return target


def make_basic_gaussian(
    N=50,
    M=50,
    pixelscale=0.8,
    x=24.5,
    y=25.4,
    PA=45 * np.pi / 180,
    sigma=3,
    flux=1,
    rand=12345,
):

    np.random.seed(rand)
    target = ap.image.Target_Image(
        data=np.zeros((N, M)),
        pixelscale=pixelscale,
        psf=ap.utils.initialize.gaussian_psf(2 / pixelscale, 11, pixelscale),
    )

    MODEL = ap.models.Gaussian_Galaxy(
        name="basic gaussian source",
        target=target,
        parameters={
            "center": [x, y],
            "sigma": sigma,
            "flux": flux,
            "PA": {"value": 0., "locked": True},
            "q": {"value": 0.99, "locked": True},
        },
    )

    img = MODEL().data.detach().cpu().numpy()
    target.data = (
        img
        + np.random.normal(scale=0.1, size=img.shape)
        + np.random.normal(scale=np.sqrt(img) / 10)
    )
    target.variance = 0.1 ** 2 + img / 100

    return target


def make_basic_gaussian_psf(
    N=25,
    pixelscale=0.8,
    sigma=3,
    rand=12345,
):

    np.random.seed(rand)
    psf = ap.utils.initialize.gaussian_psf(sigma / pixelscale, N, pixelscale)
    psf += np.random.normal(scale = psf / 2)
    psf[psf < 0] = 0
    target = ap.image.PSF_Image(
        data=psf / np.sum(psf),
        pixelscale=pixelscale,
    )

    return target



from typing import Iterable


def to_hdf5_has_None(l):
    for i in range(len(l)):
        if hasattr(l[i], "__iter__") and not isinstance(l[i], str):
            l[i] = to_hdf5_has_None(l[i])
        elif l[i] is None:
            return True
    return False


def dict_to_hdf5(h, D):
    for key in D:
        if isinstance(D[key], dict):
            n = h.create_group(key)
            dict_to_hdf5(n, D[key])
        else:
            if hasattr(D[key], "__iter__") and not isinstance(D[key], str):
                if to_hdf5_has_None(D[key]):
                    h[key] = str(D[key])
                else:
                    h.create_dataset(key, data=D[key])
            elif D[key] is not None:
                h[key] = D[key]
            else:
                h[key] = "None"


def hdf5_to_dict(h):
    import h5py

    D = {}
    for key in h.keys():
        if isinstance(h[key], h5py.Group):
            D[key] = hdf5_to_dict(h[key])
        elif isinstance(h[key], str) and "None" in h[key]:
            D[key] = eval(h[key])
        else:
            D[key] = h[key]
    return D



from typing import Optional

from scipy.stats import iqr
import torch
import numpy as np

from .model_object import Component_Model
from ._shared_methods import select_target
from ..utils.initialize import isophotes
from ..utils.angle_operations import Angle_Average
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..param import Param_Unlock, Param_SoftLimits, Parameter_Node
from ..image import Image
from ..utils.conversions.coordinates import (
    Rotate_Cartesian,
    Axis_Ratio_Cartesian,
)

__all__ = ["Edgeon_Model"]


class Edgeon_Model(Component_Model):
    """General Edge-On galaxy model to be subclassed for any specific
    representation such as radial light profile or the structure of
    the galaxy on the sky. Defines an edgeon galaxy as an object with
    a position angle, no inclination information is included.

    """

    model_type = f"edgeon {Component_Model.model_type}"
    parameter_specs = {
        "PA": {
            "units": "rad",
            "limits": (0, np.pi),
            "cyclic": True,
            "uncertainty": 0.06,
        },
    }
    _parameter_order = Component_Model._parameter_order + ("PA",)
    useable = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(
        self, target=None, parameters: Optional[Parameter_Node] = None, **kwargs
    ):
        super().initialize(target=target, parameters=parameters)
        if parameters["PA"].value is not None:
            return
        target_area = target[self.window]
        edge = np.concatenate(
            (
                target_area.data[:, 0].detach().cpu().numpy(),
                target_area.data[:, -1].detach().cpu().numpy(),
                target_area.data[0, :].detach().cpu().numpy(),
                target_area.data[-1, :].detach().cpu().numpy(),
            )
        )
        edge_average = np.median(edge)
        edge_scatter = iqr(edge, rng=(16, 84)) / 2
        icenter = target_area.plane_to_pixel(parameters["center"].value)

        iso_info = isophotes(
            target_area.data.detach().cpu().numpy() - edge_average,
            (icenter[1].detach().cpu().item(), icenter[0].detach().cpu().item()),
            threshold=3 * edge_scatter,
            pa=0.0,
            q=1.0,
            n_isophotes=15,
        )
        with Param_Unlock(parameters["PA"]), Param_SoftLimits(parameters["PA"]):
            parameters["PA"].value = (
                -(
                    (
                        Angle_Average(
                            list(
                                iso["phase2"]
                                for iso in iso_info[-int(len(iso_info) / 3) :]
                            )
                        )
                        / 2
                    )
                    + target.north
                )
            ) % np.pi
            parameters["PA"].uncertainty = parameters["PA"].value * self.default_uncertainty

    @default_internal
    def transform_coordinates(self, X, Y, image=None, parameters=None):
        return Rotate_Cartesian(-(parameters["PA"].value - image.north), X, Y)

    @default_internal
    def evaluate_model(
        self,
        X=None,
        Y=None,
        image: Image = None,
        parameters: Parameter_Node = None,
        **kwargs,
    ):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]
        XX, YY = self.transform_coordinates(X, Y, image=image, parameters=parameters)

        return self.brightness_model(
            torch.abs(XX), torch.abs(YY), image=image, parameters=parameters
        )


class Edgeon_Sech(Edgeon_Model):
    """An edgeon profile where the vertical distribution is a sech^2
    profile, subclasses define the radial profile.

    """

    model_type = f"sech2 {Edgeon_Model.model_type}"
    parameter_specs = {
        "I0": {"units": "log10(flux/arcsec^2)"},
        "hs": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Edgeon_Model._parameter_order + ("I0", "hs")
    useable = False

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(
        self, target=None, parameters: Optional[Parameter_Node] = None, **kwargs
    ):
        super().initialize(target=target, parameters=parameters)
        if (parameters["I0"].value is not None) and (
            parameters["hs"].value is not None
        ):
            return
        target_area = target[self.window]
        icenter = target_area.plane_to_pixel(parameters["center"].value)

        if parameters["I0"].value is None:
            with Param_Unlock(parameters["I0"]), Param_SoftLimits(parameters["I0"]):
                parameters["I0"].value = torch.log10(
                    torch.mean(
                        target_area.data[
                            int(icenter[0]) - 2 : int(icenter[0]) + 2,
                            int(icenter[1]) - 2 : int(icenter[1]) + 2,
                        ]
                    )
                    / target.pixel_area.item()
                )
                parameters["I0"].uncertainty = torch.std(
                    target_area.data[
                        int(icenter[0]) - 2 : int(icenter[0]) + 2,
                        int(icenter[1]) - 2 : int(icenter[1]) + 2,
                    ]
                ) / (torch.abs(parameters["I0"].value) * target.pixel_area)
        if parameters["hs"].value is None:
            with Param_Unlock(parameters["hs"]), Param_SoftLimits(parameters["hs"]):
                parameters["hs"].value = torch.max(self.window.shape) * 0.1
                parameters["hs"].uncertainty = parameters["hs"].value / 2

    @default_internal
    def brightness_model(self, X, Y, image=None, parameters=None):
        return (
            (image.pixel_area * 10 ** parameters["I0"].value)
            * self.radial_model(X, image=image, parameters=parameters)
            / (torch.cosh((Y + self.softening) / parameters["hs"].value) ** 2)
        )


class Edgeon_Isothermal(Edgeon_Sech):
    """A self-gravitating locally-isothermal edgeon disk. This comes from
    van der Kruit & Searle 1981.

    """

    model_type = f"isothermal {Edgeon_Sech.model_type}"
    parameter_specs = {
        "rs": {"units": "arcsec", "limits": (0, None)},
    }
    _parameter_order = Edgeon_Sech._parameter_order + ("rs",)
    useable = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(
        self, target=None, parameters: Optional[Parameter_Node] = None, **kwargs
    ):
        super().initialize(target=target, parameters=parameters)
        if parameters["rs"].value is not None:
            return
        with Param_Unlock(parameters["rs"]), Param_SoftLimits(parameters["rs"]):
            parameters["rs"].value = torch.max(self.window.shape) * 0.4
            parameters["rs"].uncertainty = parameters["rs"].value / 2

    @default_internal
    def radial_model(self, R, image=None, parameters=None):
        Rscaled = torch.abs((R + self.softening) / parameters["rs"].value)
        return (
            Rscaled
            * torch.exp(-Rscaled)
            * torch.special.scaled_modified_bessel_k1(Rscaled)
        )



import torch

from .psf_model_object import PSF_Model
from ..image import PSF_Image
from ..utils.decorators import ignore_numpy_warnings, default_internal
from ..utils.interpolate import interp2d
from ._shared_methods import select_target
from ..param import Param_Unlock, Param_SoftLimits
from .. import AP_config

__all__ = ["Pixelated_PSF"]


class Pixelated_PSF(PSF_Model):
    """point source model which uses an image of the PSF as its
    representation for point sources. Using bilinear interpolation it
    will shift the PSF within a pixel to accurately represent the
    center location of a point source. There is no funcitonal form for
    this object type as any image can be supplied. The image pixels
    will be optimized as individual parameters. This can very quickly
    result in a large number of parameters and a near impossible
    fitting task, ideally this should be restricted to a very small
    area likely at the center of the PSF.

    To initialize the PSF image will by default be set to the target
    PSF_Image values, thus one can use an empirical PSF as a starting
    point. Since only bilinear interpolation is performed, it is
    recommended to provide the PSF at a higher resolution than the
    image if it is near the nyquist sampling limit. Bilinear
    interpolation is very fast and accurate for smooth models, so this
    way it is possible to do the expensive interpolation before
    optimization and save time. Note that if you do this you must
    provide the PSF as a PSF_Image object with the correct pixelscale
    (essentially just divide the pixelscale by the upsampling factor
    you used).

    Parameters:
        pixels: the total flux within each pixel, represented as the log of the flux.

    """

    model_type = f"pixelated {PSF_Model.model_type}"
    parameter_specs = {
        "pixels": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = PSF_Model._parameter_order + ("pixels",)
    useable = True
    model_integrated = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)
        target_area = target[self.window]
        with Param_Unlock(parameters["pixels"]), Param_SoftLimits(parameters["pixels"]):
            if parameters["pixels"].value is None:
                dat = torch.abs(target_area.data)
                dat[dat == 0] = torch.median(dat) * 1e-7
                parameters["pixels"].value = torch.log10(dat / target.pixel_area)
            if parameters["pixels"].uncertainty is None:
                parameters["pixels"].uncertainty = torch.abs(parameters["pixels"].value) * self.default_uncertainty

    @default_internal
    def evaluate_model(self, X=None, Y=None, image=None, parameters=None, **kwargs):
        if X is None:
            Coords = image.get_coordinate_meshgrid()
            X, Y = Coords - parameters["center"].value[..., None, None]

        # Convert coordinates into pixel locations in the psf image
        pX, pY = self.target.plane_to_pixel(X, Y)

        # Select only the pixels where the PSF image is defined
        select = torch.logical_and(
            torch.logical_and(pX > -0.5, pX < parameters["pixels"].shape[1] - 0.5),
            torch.logical_and(pY > -0.5, pY < parameters["pixels"].shape[0] - 0.5),
        )

        # Zero everywhere outside the psf
        result = torch.zeros_like(X)

        # Use bilinear interpolation of the PSF at the requested coordinates
        result[select] = interp2d(parameters["pixels"].value, pX[select], pY[select])

        return image.pixel_area * 10 ** result



# Metropolis-Hasting Markov-Chain Monte-Carlo
import os
from time import time
from typing import Optional, Sequence
import torch
from tqdm import tqdm
import numpy as np
from .base import BaseOptimizer
from .. import AP_config

__all__ = ["MHMCMC"]


class MHMCMC(BaseOptimizer):
    """Metropolis-Hastings Markov-Chain Monte-Carlo sampler, based on:
    https://en.wikipedia.org/wiki/Metropolis-Hastings_algorithm . This
    is a naive implimentation of a standard MCMC, it is far from
    optimal and should not be used for anything but the most basic
    scenarios.

    Args:
      model (AstroPhot_Model): The model which will be sampled.
      initial_state (Optional[Sequence]): A 1D array with the values for each parameter in the model. Note that these values should be in the form of "as_representation" in the model.
      max_iter (int): The number of sampling steps to perform. Default 1000
      epsilon (float or array): The random step length to take at each iteration. This is the standard deviation for the normal distribution sampling. Default 1e-2

    """

    def __init__(
        self,
        model: "AstroPhot_Model",
        initial_state: Optional[Sequence] = None,
        max_iter: int = 1000,
        **kwargs,
    ):
        super().__init__(model, initial_state, max_iter=max_iter, **kwargs)

        self.epsilon = kwargs.get("epsilon", 1e-2)
        self.progress_bar = kwargs.get("progress_bar", True)
        self.report_after = kwargs.get("report_after", int(self.max_iter / 10))

        self.chain = []
        self._accepted = 0
        self._sampled = 0

    def fit(
        self,
        state: Optional[torch.Tensor] = None,
        nsamples: Optional[int] = None,
        restart_chain: bool = True,
    ):
        """
        Performs the MCMC sampling using a Metropolis Hastings acceptance step and records the chain for later examination.
        """

        if nsamples is None:
            nsamples = self.max_iter

        if state is None:
            state = self.current_state
        chi2 = self.sample(state)

        if restart_chain:
            self.chain = []
        else:
            self.chain = list(self.chain)

        iterator = tqdm(range(nsamples)) if self.progress_bar else range(nsamples)
        for i in iterator:
            state, chi2 = self.step(state, chi2)
            self.append_chain(state)
            if i % self.report_after == 0 and i > 0 and self.verbose > 0:
                AP_config.ap_logger.info(f"Acceptance: {self.acceptance}")
        if self.verbose > 0:
            AP_config.ap_logger.info(f"Acceptance: {self.acceptance}")
        self.current_state = state
        self.chain = np.stack(self.chain)
        return self

    def append_chain(self, state: torch.Tensor):
        """
        Add a state vector to the MCMC chain
        """

        self.chain.append(
            self.model.parameters.vector_transform_rep_to_val(state)
            .detach()
            .cpu()
            .clone()
            .numpy()
        )

    @staticmethod
    def accept(log_alpha):
        """
        Evaluates randomly if a given proposal is accepted. This is done in log space which is more natural for the evaluation in the step.
        """
        return torch.log(torch.rand(log_alpha.shape)) < log_alpha

    @torch.no_grad()
    def sample(self, state: torch.Tensor):
        """
        Samples the model at the proposed state vector values
        """
        return self.model.negative_log_likelihood(
            parameters=state, as_representation=True
        )

    @torch.no_grad()
    def step(self, state: torch.Tensor, chi2: torch.Tensor) -> torch.Tensor:
        """
        Takes one step of the HMC sampler by integrating along a path initiated with a random momentum.
        """

        proposal_state = torch.normal(mean=state, std=self.epsilon)
        proposal_chi2 = self.sample(proposal_state)
        log_alpha = chi2 - proposal_chi2
        accept = self.accept(log_alpha)
        self._accepted += accept
        self._sampled += 1
        return proposal_state if accept else state, proposal_chi2 if accept else chi2

    @property
    def acceptance(self):
        """
        Returns the ratio of accepted states to total states sampled.
        """

        return self._accepted / self._sampled



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import astrophot as ap

# Add the package root to the system path to enable autodoc to find modules.
sys.path.insert(0, os.path.abspath("../"))

# -- General configuration ------------------------------------------------

# Extensions to use
extensions = [
    'nbsphinx',
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    'sphinx.ext.viewcode',
]

# Paths to templates
templates_path = ["_templates"]

# Suffixes of source filenames
source_suffix = ".rst"

# Master document
master_doc = "index"

# Project information
project = "AstroPhot"
copyright = "2023, Connor Stone"
author = "Connor Stone"

# Version information
version = ap.__version__
release = ap.__version__

# Patterns of files and directories to exclude
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# Pygments style to use
pygments_style = "sphinx"

# Whether to include TODOs
todo_include_todos = False

# -- Options for HTML output ----------------------------------------------

# Theme to use
html_theme = "sphinx_rtd_theme"

# Sidebar templates
html_sidebars = {
    "**": [
        "relations.html",
        "searchbox.html",
    ]
}

html_favicon = "media/AP_logo_favicon.ico"

# Output file base name for HTML help builder
htmlhelp_basename = "AstroPhotdocs"

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {}

# LaTeX document settings
latex_documents = [
    (master_doc, "AstroPhot.tex", "AstroPhot Documentation", "Connor Stone", "manual"),
]

# -- Options for manual page output ---------------------------------------

man_pages = [
    (master_doc, "astrophot", "AstroPhot Documentation", [author], 1),
]

# -- Options for Texinfo output -------------------------------------------

texinfo_documents = [
    (
        master_doc,
        "AstroPhot",
        "AstroPhot Documentation",
        author,
        "AstroPhot",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# -- Options for nbsphinx --------------------------------------------------
nbsphinx_execute = 'never'



import unittest
from astrophot import image
import astrophot as ap
import torch
import numpy as np

from utils import get_astropy_wcs

######################################################################
# Image Objects
######################################################################


class TestImage(unittest.TestCase):
    def test_image_creation(self):
        arr = torch.zeros((10, 15))
        base_image = image.Image(
            data=arr,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
            metadata={"note": "test image"},
        )

        self.assertEqual(base_image.pixel_length, 1.0, "image should track pixelscale")
        self.assertEqual(base_image.zeropoint, 1.0, "image should track zeropoint")
        self.assertEqual(base_image.origin[0], 0, "image should track origin")
        self.assertEqual(base_image.origin[1], 0, "image should track origin")
        self.assertEqual(base_image.metadata["note"], "test image", "image should track note")

        slicer = image.Window(origin=(3, 2), pixel_shape=(4, 5))
        sliced_image = base_image[slicer]
        self.assertEqual(sliced_image.origin[0], 3, "image should track origin")
        self.assertEqual(sliced_image.origin[1], 2, "image should track origin")
        self.assertEqual(base_image.origin[0], 0, "subimage should not change image origin")
        self.assertEqual(base_image.origin[1], 0, "subimage should not change image origin")

        second_base_image = image.Image(data=arr, pixelscale=1.0, metadata={"note": "test image"})
        self.assertEqual(base_image.pixel_length, 1.0, "image should track pixelscale")
        self.assertIsNone(second_base_image.zeropoint, "image should track zeropoint")
        self.assertEqual(second_base_image.origin[0], 0, "image should track origin")
        self.assertEqual(second_base_image.origin[1], 0, "image should track origin")
        self.assertEqual(
            second_base_image.metadata["note"], "test image", "image should track note"
        )

    def test_copy(self):

        new_image = image.Image(
            data=torch.zeros((10, 15)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )

        copy_image = new_image.copy()
        self.assertEqual(
            new_image.pixel_length,
            copy_image.pixel_length,
            "copied image should have same pixelscale",
        )
        self.assertEqual(
            new_image.zeropoint,
            copy_image.zeropoint,
            "copied image should have same zeropoint",
        )
        self.assertEqual(
            new_image.window, copy_image.window, "copied image should have same window"
        )
        copy_image += 1
        self.assertEqual(
            new_image.data[0][0],
            0.0,
            "copied image should not share data with original",
        )

        blank_copy_image = new_image.blank_copy()
        self.assertEqual(
            new_image.pixel_length,
            blank_copy_image.pixel_length,
            "copied image should have same pixelscale",
        )
        self.assertEqual(
            new_image.zeropoint,
            blank_copy_image.zeropoint,
            "copied image should have same zeropoint",
        )
        self.assertEqual(
            new_image.window,
            blank_copy_image.window,
            "copied image should have same window",
        )
        blank_copy_image += 1
        self.assertEqual(
            new_image.data[0][0],
            0.0,
            "copied image should not share data with original",
        )

    def test_image_arithmetic(self):

        arr = torch.zeros((10, 12))
        base_image = image.Image(
            data=arr,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.ones(2),
        )
        slicer = image.Window(origin=(0, 0), pixel_shape=(5, 5))
        sliced_image = base_image[slicer]
        sliced_image += 1

        self.assertEqual(base_image.data[1][1], 1, "slice should update base image")
        self.assertEqual(base_image.data[5][5], 0, "slice should only update its region")

        second_image = image.Image(
            data=torch.ones((5, 5)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=[3, 3],
        )

        # Test iadd
        base_image += second_image
        self.assertEqual(base_image.data[1][1], 1, "image addition should only update its region")
        self.assertEqual(base_image.data[3][3], 2, "image addition should update its region")
        self.assertEqual(base_image.data[5][5], 1, "image addition should update its region")
        self.assertEqual(base_image.data[8][8], 0, "image addition should only update its region")

        # Test isubtract
        base_image -= second_image
        self.assertEqual(
            base_image.data[1][1], 1, "image subtraction should only update its region"
        )
        self.assertEqual(base_image.data[3][3], 1, "image subtraction should update its region")
        self.assertEqual(base_image.data[5][5], 0, "image subtraction should update its region")
        self.assertEqual(
            base_image.data[8][8], 0, "image subtraction should only update its region"
        )

        base_image.data[6:, 6:] += 1.0

        self.assertEqual(base_image.data[1][1], 1, "array addition should only update its region")
        self.assertEqual(base_image.data[6][6], 1, "array addition should update its region")
        self.assertEqual(base_image.data[8][8], 1, "array addition should update its region")

    def test_excersize_arithmatic(self):

        arr = torch.zeros((10, 12))
        base_image = image.Image(
            data=arr,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.ones(2),
        )
        second_image = image.Image(
            data=torch.ones((5, 5)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=[3, 3],
        )

        new_img = base_image + second_image
        new_img = new_img - second_image

        self.assertTrue(
            torch.allclose(new_img.data, torch.zeros_like(new_img.data)),
            "addition and subtraction should produce no change",
        )

        base_image += second_image
        base_image -= second_image

        self.assertTrue(
            torch.allclose(base_image.data, torch.zeros_like(base_image.data)),
            "addition and subtraction should produce no change",
        )

        new_img = base_image + 10.0
        new_img = new_img - 10.0

        self.assertTrue(
            torch.allclose(new_img.data, torch.zeros_like(new_img.data)),
            "addition and subtraction should produce no change",
        )

        base_image += 10.0
        base_image -= 10.0

        self.assertTrue(
            torch.allclose(base_image.data, torch.zeros_like(base_image.data)),
            "addition and subtraction should produce no change",
        )

    def test_image_manipulation(self):

        new_image = image.Image(
            data=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )

        # image reduction
        for scale in [2, 4, 8, 16]:
            reduced_image = new_image.reduce(scale)

            self.assertEqual(
                reduced_image.data[0][0],
                scale**2,
                "reduced image should sum sub pixels",
            )
            self.assertEqual(
                reduced_image.pixel_length,
                scale,
                "pixelscale should increase with reduced image",
            )
            self.assertEqual(
                reduced_image.origin[0],
                new_image.origin[0],
                "origin should not change with reduced image",
            )
            self.assertEqual(
                reduced_image.shape[0],
                new_image.shape[0],
                "shape should not change with reduced image",
            )

        # iamge cropping
        new_image.crop(
            [torch.tensor(1, dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device)]
        )
        self.assertEqual(
            new_image.data.shape[0], 14, "crop should cut 1 pixel from both sides here"
        )
        new_image.crop(
            torch.tensor([3, 2], dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device)
        )
        self.assertEqual(
            new_image.data.shape[1],
            24,
            "previous crop and current crop should have cut from this axis",
        )
        new_image.crop(
            torch.tensor([3, 2, 1, 0], dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device)
        )
        self.assertEqual(
            new_image.data.shape[0],
            9,
            "previous crop and current crop should have cut from this axis",
        )

    def test_image_save_load(self):

        new_image = image.Image(
            data=torch.ones((16, 32)),
            pixelscale=0.76,
            zeropoint=21.4,
            origin=torch.zeros(2) + 0.1,
        )

        new_image.save("Test_AstroPhot.fits")

        loaded_image = ap.image.Image(filename="Test_AstroPhot.fits")

        self.assertTrue(
            torch.all(new_image.data == loaded_image.data),
            "Loaded image should have same pixel values",
        )
        self.assertTrue(
            torch.all(new_image.origin == loaded_image.origin),
            "Loaded image should have same origin",
        )
        self.assertEqual(
            new_image.pixel_length,
            loaded_image.pixel_length,
            "Loaded image should have same pixel scale",
        )
        self.assertEqual(
            new_image.zeropoint,
            loaded_image.zeropoint,
            "Loaded image should have same zeropoint",
        )

    def test_image_wcs_roundtrip(self):

        wcs = get_astropy_wcs()
        # Minimial input
        I = ap.image.Image(
            data=torch.zeros((20, 20)),
            zeropoint=22.5,
            wcs=wcs,
        )

        self.assertTrue(
            torch.allclose(
                I.world_to_plane(I.plane_to_world(torch.zeros_like(I.window.reference_radec))),
                torch.zeros_like(I.window.reference_radec),
            ),
            "WCS world/plane roundtrip should return input value",
        )
        self.assertTrue(
            torch.allclose(
                I.pixel_to_plane(I.plane_to_pixel(torch.zeros_like(I.window.reference_radec))),
                torch.zeros_like(I.window.reference_radec),
            ),
            "WCS pixel/plane roundtrip should return input value",
        )
        self.assertTrue(
            torch.allclose(
                I.world_to_pixel(I.pixel_to_world(torch.zeros_like(I.window.reference_radec))),
                torch.zeros_like(I.window.reference_radec),
                atol=1e-6,
            ),
            "WCS world/pixel roundtrip should return input value",
        )

        self.assertTrue(
            torch.allclose(
                I.pixel_to_plane_delta(
                    I.plane_to_pixel_delta(torch.ones_like(I.window.reference_radec))
                ),
                torch.ones_like(I.window.reference_radec),
            ),
            "WCS pixel/plane delta roundtrip should return input value",
        )

    def test_image_display(self):
        new_image = image.Image(
            data=torch.ones((16, 32)),
            pixelscale=0.76,
            zeropoint=21.4,
            origin=torch.zeros(2) + 0.1,
        )

        self.assertIsInstance(str(new_image), str, "String representation should be a string!")
        self.assertIsInstance(repr(new_image), str, "Repr should be a string!")

    def test_image_errors(self):

        new_image = image.Image(
            data=torch.ones((16, 32)),
            pixelscale=0.76,
            zeropoint=21.4,
            origin=torch.zeros(2) + 0.1,
        )

        # Change data badly
        with self.assertRaises(ap.errors.SpecificationConflict):
            new_image.data = np.zeros((5, 5))

        # Fractional image reduction
        with self.assertRaises(ap.errors.SpecificationConflict):
            reduced = new_image.reduce(0.2)

        # Negative expand image
        with self.assertRaises(ap.errors.SpecificationConflict):
            unexpanded = new_image.expand((-2, 3))


class TestTargetImage(unittest.TestCase):
    def test_variance(self):

        new_image = image.Target_Image(
            data=torch.ones((16, 32)),
            variance=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )

        self.assertTrue(new_image.has_variance, "target image should store variance")

        reduced_image = new_image.reduce(2)
        self.assertEqual(reduced_image.variance[0][0], 4, "reduced image should sum sub pixels")

        new_image.to()
        new_image.variance = None
        self.assertFalse(new_image.has_variance, "target image update to no variance")

    def test_mask(self):

        new_image = image.Target_Image(
            data=torch.ones((16, 32)),
            mask=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        self.assertTrue(new_image.has_mask, "target image should store mask")

        reduced_image = new_image.reduce(2)
        self.assertEqual(reduced_image.mask[0][0], 1, "reduced image should mask apropriately")

        new_image.mask = None
        self.assertFalse(new_image.has_mask, "target image update to no mask")

        data = torch.ones((16, 32))
        data[1, 1] = torch.nan
        data[5, 5] = torch.nan

        new_image = image.Target_Image(
            data=data,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        self.assertTrue(new_image.has_mask, "target image with nans should create mask")
        self.assertEqual(new_image.mask[1][1].item(), True, "nan should be masked")
        self.assertEqual(new_image.mask[5][5].item(), True, "nan should be masked")

    def test_psf(self):

        new_image = image.Target_Image(
            data=torch.ones((15, 33)),
            psf=torch.ones((9, 9)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        self.assertTrue(new_image.has_psf, "target image should store variance")
        self.assertEqual(
            new_image.psf.psf_border_int[0],
            5,
            "psf border should be half psf size, rounded up ",
        )

        reduced_image = new_image.reduce(3)
        self.assertEqual(
            reduced_image.psf.data[0][0], 9, "reduced image should sum sub pixels in psf"
        )

        new_image.psf = None
        self.assertFalse(new_image.has_psf, "target image update to no variance")

    def test_reduce(self):
        new_image = image.Target_Image(
            data=torch.ones((30, 36)),
            psf=torch.ones((9, 9)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        smaller_image = new_image.reduce(3)
        self.assertEqual(smaller_image.data[0][0], 9, "reduction should sum flux")
        self.assertEqual(
            tuple(smaller_image.data.shape), (10, 12), "reduction should decrease image size"
        )
        self.assertEqual(smaller_image.psf.data[0][0], 9, "reduction should sum psf flux")
        self.assertEqual(
            tuple(smaller_image.psf.data.shape), (3, 3), "reduction should decrease psf image size"
        )

    def test_target_save_load(self):
        new_image = image.Target_Image(
            data=torch.ones((16, 32)),
            variance=torch.ones((16, 32)),
            mask=torch.zeros((16, 32)),
            psf=torch.ones((9, 9)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )

        new_image.save("Test_target_AstroPhot.fits")

        loaded_image = ap.image.Target_Image(filename="Test_target_AstroPhot.fits")

        self.assertTrue(
            torch.all(new_image.variance == loaded_image.variance),
            "Loaded image should have same variance",
        )
        self.assertTrue(
            torch.all(new_image.psf.data == loaded_image.psf.data),
            "Loaded image should have same psf",
        )

    def test_target_errors(self):
        new_image = image.Target_Image(
            data=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )

        # bad variance
        with self.assertRaises(ap.errors.SpecificationConflict):
            new_image.variance = np.ones((5, 5))

        # bad mask
        with self.assertRaises(ap.errors.SpecificationConflict):
            new_image.mask = np.zeros((5, 5))


class TestPSFImage(unittest.TestCase):
    def test_copying(self):
        psf_image = image.PSF_Image(
            data=torch.ones((15, 15)),
            pixelscale=1.0,
        )

        copy_psf = psf_image.copy()
        self.assertEqual(
            psf_image.data[0][0], copy_psf.data[0][0], "copied image should have same data"
        )
        blank_psf = psf_image.blank_copy()
        self.assertNotEqual(
            psf_image.data[0][0],
            blank_psf.data[0][0],
            "blank copied image should not have same data",
        )

        psf_image.to(dtype=torch.float32)

    def test_reducing(self):
        psf_image = image.PSF_Image(
            data=torch.ones((15, 15)),
            pixelscale=1.0,
        )
        new_image = image.Target_Image(
            data=torch.ones((36, 45)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
            psf=psf_image,
        )

        reduce_image = new_image.reduce(3)
        self.assertEqual(
            tuple(reduce_image.psf.data.shape), (5, 5), "reducing image should reduce psf"
        )
        self.assertEqual(
            reduce_image.psf.pixel_length, 3, "reducing image should update pixelscale factor"
        )

    def test_psf_errors(self):
        with self.assertRaises(ap.errors.SpecificationConflict):
            psf_image = image.PSF_Image(
                data=torch.ones((18, 15)),
                pixelscale=1.0,
            )


class TestModelImage(unittest.TestCase):
    def test_replace(self):
        new_image = image.Model_Image(
            data=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        other_image = image.Model_Image(
            data=5 * torch.ones((4, 4)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 4 + 0.1,
        )

        new_image.replace(other_image)
        new_image.replace(other_image.window, other_image.data)

        self.assertEqual(
            new_image.data[0][0],
            1,
            "image replace should occur at proper location in image, this data should be untouched",
        )
        self.assertEqual(
            new_image.data[5][5], 5, "image replace should update values in its window"
        )

    def test_shift(self):

        new_image = image.Model_Image(
            data=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        new_image.shift_origin(
            torch.tensor((-0.1, -0.1), dtype=ap.AP_config.ap_dtype, device=ap.AP_config.ap_device),
            is_prepadded=False,
        )

        self.assertAlmostEqual(
            torch.sum(new_image.data).item(),
            16 * 32,
            delta=1,
            msg="Shifting field of ones should give field of ones",
        )

    def test_errors(self):

        with self.assertRaises(ap.errors.InvalidData):
            new_image = image.Model_Image()


class TestJacobianImage(unittest.TestCase):
    def test_jacobian_add(self):

        new_image = ap.image.Jacobian_Image(
            parameters=["a", "b", "c"],
            target_identity="target1",
            data=torch.ones((16, 32, 3)),
            pixelscale=1.0,
            zeropoint=1.0,
            window=ap.image.Window(origin=torch.zeros(2) + 0.1, pixel_shape=torch.tensor((32, 16))),
        )
        other_image = ap.image.Jacobian_Image(
            parameters=["b", "d"],
            target_identity="target1",
            data=5 * torch.ones((4, 4, 2)),
            pixelscale=1.0,
            zeropoint=1.0,
            window=ap.image.Window(
                origin=torch.zeros(2) + 4 + 0.1, pixel_shape=torch.tensor((4, 4))
            ),
        )

        new_image += other_image

        self.assertEqual(
            tuple(new_image.data.shape),
            (16, 32, 4),
            "Jacobian addition should manage parameter identities",
        )
        self.assertEqual(
            tuple(new_image.flatten("data").shape),
            (512, 4),
            "Jacobian should flatten to Npix*Nparams tensor",
        )

    def test_jacobian_error(self):

        # Create parameter list with multiple same entries
        with self.assertRaises(ap.errors.SpecificationConflict):
            new_image = ap.image.Jacobian_Image(
                parameters=["a", "b", "c", "a"],
                target_identity="target1",
                data=torch.ones((16, 32, 3)),
                pixelscale=1.0,
                zeropoint=1.0,
                window=ap.image.Window(
                    origin=torch.zeros(2) + 0.1, pixel_shape=torch.tensor((32, 16))
                ),
            )

        # Adding a model image to a jacobian image
        new_image = ap.image.Jacobian_Image(
            parameters=["a", "b", "c"],
            target_identity="target1",
            data=torch.ones((16, 32, 3)),
            pixelscale=1.0,
            zeropoint=1.0,
            window=ap.image.Window(origin=torch.zeros(2) + 0.1, pixel_shape=torch.tensor((32, 16))),
        )
        bad_image = image.Model_Image(
            data=torch.ones((16, 32)),
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2) + 0.1,
        )
        with self.assertRaises(ap.errors.InvalidImage):
            new_image += bad_image


if __name__ == "__main__":
    unittest.main()



import unittest
import astrophot as ap
import numpy as np
import torch

class TestWPCS(unittest.TestCase):
    def test_wpcs_creation(self):

        # Blank startup
        wcs_blank = ap.image.WPCS()

        self.assertEqual(wcs_blank.projection, "gnomonic", "Default projection should be Gnomonic")
        self.assertTrue(torch.all(wcs_blank.reference_radec == 0), "default reference world coordinates should be zeros")
        self.assertTrue(torch.all(wcs_blank.reference_planexy == 0), "default reference plane coordinates should be zeros")

        # Provided parameters
        wcs_set = ap.image.WPCS(
            projection = "orthographic",
            reference_radec = (90,10),
        )

        self.assertEqual(wcs_set.projection, "orthographic", "Provided projection was Orthographic")
        self.assertTrue(torch.all(wcs_set.reference_radec == torch.tensor((90,10), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "World coordinates should be as provided")
        self.assertNotEqual(wcs_blank.projection, "orthographic", "Not all WCS objects should be updated")
        self.assertFalse(torch.all(wcs_blank.reference_radec == torch.tensor((90,10), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "Not all WCS objects should be updated")

        wcs_set = wcs_set.copy()
        
        self.assertEqual(wcs_set.projection, "orthographic", "Provided projection was Orthographic")
        self.assertTrue(torch.all(wcs_set.reference_radec == torch.tensor((90,10), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "World coordinates should be as provided")
        self.assertNotEqual(wcs_blank.projection, "orthographic", "Not all WCS objects should be updated")
        self.assertFalse(torch.all(wcs_blank.reference_radec == torch.tensor((90,10), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "Not all WCS objects should be updated")

    def test_wpcs_round_trip(self):

        for projection in ["gnomonic", "orthographic", "steriographic"]:
            print(projection)
            for ref_coords in [(20.3, 79), (120.2,-19), (300, -50), (0,0)]:
                print(ref_coords)
                wcs = ap.image.WPCS(
                    projection = projection,
                    reference_radec = ref_coords,
                )

                test_grid_RA, test_grid_DEC = torch.meshgrid(
                    torch.linspace(ref_coords[0] - 10, ref_coords[0] + 10, 10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), # RA
                    torch.linspace(ref_coords[1] - 10, ref_coords[1] + 10, 10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), # DEC
                    indexing = "xy",
                )

                project_x, project_y = wcs.world_to_plane(
                    test_grid_RA,
                    test_grid_DEC,
                )

                reproject_RA, reproject_DEC = wcs.plane_to_world(
                    project_x,
                    project_y,
                )

                self.assertTrue(torch.allclose(reproject_RA, test_grid_RA), "Round trip RA should map back to itself")
                self.assertTrue(torch.allclose(reproject_DEC, test_grid_DEC), "Round trip DEC should map back to itself")

    def test_wpcs_errors(self):
        with self.assertRaises(ap.errors.InvalidWCS):
            wcs = ap.image.WPCS(
                projection = "connor",
            )
        
class TestPPCS(unittest.TestCase):

    def test_ppcs_creation(self):
        # Blank startup
        wcs_blank = ap.image.PPCS()

        self.assertTrue(np.all(wcs_blank.pixelscale.detach().cpu().numpy() == np.array([[1.,0.],[0.,1.]])), "Default pixelscale should be 1")
        self.assertTrue(torch.all(wcs_blank.reference_imageij == -0.5), "default reference pixel coordinates should be -0.5")
        self.assertTrue(torch.all(wcs_blank.reference_imagexy == 0.), "default reference plane coordinates should be zeros")

        # Provided parameters
        wcs_set = ap.image.PPCS(
            pixelscale = [[-0.173205, 0.1],[0.15,0.259808]],
            reference_imageij = (5,10),
            reference_imagexy = (0.12, 0.45),
        )

        self.assertTrue(torch.allclose(wcs_set.pixelscale, torch.tensor([[-0.173205, 0.1],[0.15,0.259808]], dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "Provided pixelscale should be used")
        self.assertTrue(torch.allclose(wcs_set.reference_imageij, torch.tensor((5.,10.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "pixel reference coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs_set.reference_imagexy, torch.tensor((0.12,0.45), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "plane reference coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs_set.plane_to_pixel(torch.tensor((0.12,0.45), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), torch.tensor((5.,10.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "plane reference coordinates should map to pixel reference coordinates")
        self.assertTrue(torch.allclose(wcs_set.pixel_to_plane(torch.tensor((5.,10.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), torch.tensor((0.12,0.45), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "pixel reference coordinates should map to plane reference coordinates")

        wcs_set = wcs_set.copy()

        self.assertTrue(torch.allclose(wcs_set.pixelscale, torch.tensor([[-0.173205, 0.1],[0.15,0.259808]], dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "Provided pixelscale should be used")
        self.assertTrue(torch.allclose(wcs_set.reference_imageij, torch.tensor((5.,10.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "pixel reference coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs_set.reference_imagexy, torch.tensor((0.12,0.45), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "plane reference coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs_set.plane_to_pixel(torch.tensor((0.12,0.45), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), torch.tensor((5.,10.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "plane reference coordinates should map to pixel reference coordinates")
        self.assertTrue(torch.allclose(wcs_set.pixel_to_plane(torch.tensor((5.,10.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), torch.tensor((0.12,0.45), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "pixel reference coordinates should map to plane reference coordinates")
        

        wcs_set.pixelscale = None
        
    def test_ppcs_round_trip(self):

        for pixelscale in [0.2, [[0.6, 0.],[0., 0.4]], [[-0.173205, 0.1],[0.15,0.259808]]]:
            print(pixelscale)
            for ref_coords in [(20.3, 79), (120.2,-19), (300, -50), (0,0)]:
                print(ref_coords)
                wcs = ap.image.PPCS(
                    pixelscale = pixelscale,
                    reference_imagexy = ref_coords,
                )

                test_grid_x, test_grid_y = torch.meshgrid(
                    torch.linspace(ref_coords[0] - 10, ref_coords[0] + 10, 10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), # x
                    torch.linspace(ref_coords[1] - 10, ref_coords[1] + 10, 10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), # y
                    indexing = "xy",
                )

                project_i, project_j = wcs.plane_to_pixel(
                    test_grid_x,
                    test_grid_y,
                )

                reproject_x, reproject_y = wcs.pixel_to_plane(
                    project_i,
                    project_j,
                )

                self.assertTrue(torch.allclose(reproject_x, test_grid_x), "Round trip x should map back to itself")
                self.assertTrue(torch.allclose(reproject_y, test_grid_y), "Round trip y should map back to itself")

class TestWCS(unittest.TestCase):
    def test_wcs_creation(self):

        wcs = ap.image.WCS(
            projection = "orthographic",
            pixelscale = [[-0.173205, 0.1],[0.15,0.259808]],
            reference_radec = (120.2,-19),
            reference_imagexy = (33., 123.),
        )

        wcs2 = wcs.copy()

        self.assertEqual(wcs2.projection, "orthographic", "Provided projection was Orthographic")
        self.assertTrue(torch.allclose(wcs2.reference_radec, wcs.reference_radec), "World coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs2.reference_planexy, wcs.reference_planexy), "Plane coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs2.reference_imagexy, wcs.reference_imagexy), "imagexy coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs2.reference_imageij, wcs.reference_imageij), "imageij coordinates should be as provided")
        self.assertTrue(torch.allclose(wcs2.pixelscale, wcs.pixelscale), "pixelscale should be as provided")
        
        
    def test_wcs_roundtrip(self):
        for pixelscale in [0.2, [[0.6, 0.],[0., 0.4]], [[-0.173205, 0.1],[0.15,0.259808]]]:
            print(pixelscale)
            for ref_coords_xy in [(33., 123.), (-430.2,-11), (-97., 5), (0,0)]:
                for projection in ["gnomonic", "orthographic", "steriographic"]:
                    print(projection)
                    for ref_coords_radec in [(20.3, 79), (120.2,-19), (300, -50), (0,0)]:
                        print(ref_coords_radec)
                        wcs = ap.image.WCS(
                            projection = projection,
                            pixelscale = pixelscale,
                            reference_radec = ref_coords_radec,
                            reference_imagexy = ref_coords_xy,
                        )

                        
                        test_grid_RA, test_grid_DEC = torch.meshgrid(
                            torch.linspace(ref_coords_radec[0] - 10, ref_coords_radec[0] + 10, 10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), # RA
                            torch.linspace(ref_coords_radec[1] - 10, ref_coords_radec[1] + 10, 10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device), # DEC
                            indexing = "xy",
                        )
                        
                        project_i, project_j = wcs.world_to_pixel(
                            test_grid_RA,
                            test_grid_DEC,
                        )
                        
                        reproject_RA, reproject_DEC = wcs.pixel_to_world(
                            project_i,
                            project_j,
                        )
                        
                        self.assertTrue(torch.allclose(reproject_RA, test_grid_RA), "Round trip RA should map back to itself")
                        self.assertTrue(torch.allclose(reproject_DEC, test_grid_DEC), "Round trip DEC should map back to itself")
    

    def test_wcs_state(self):
        wcs = ap.image.WCS(
            projection = "orthographic",
            pixelscale = [[-0.173205, 0.1],[0.15,0.259808]],
            reference_radec = (120.2,-19),
            reference_imagexy = (33., 123.),
        )

        wcs_state = wcs.get_state()

        new_wcs = ap.image.WCS(state = wcs_state)

        self.assertEqual(wcs.projection, new_wcs.projection, "WCS projection should be set by state")
        self.assertTrue(torch.allclose(wcs.pixelscale, torch.tensor([[-0.173205, 0.1],[0.15,0.259808]], dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "WCS pixelscale should be set by state")
        self.assertTrue(torch.allclose(wcs.reference_radec, torch.tensor((120.2,-19), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "WCS reference RA DEC should be set by state")
        self.assertTrue(torch.allclose(wcs.reference_imagexy, torch.tensor((33., 123.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "WCS reference image position should be set by state")

        wcs_state = wcs.get_fits_state()

        new_wcs = ap.image.WCS()
        new_wcs.set_fits_state(state = wcs_state)

        self.assertEqual(wcs.projection, new_wcs.projection, "WCS projection should be set by state")
        self.assertTrue(torch.allclose(wcs.pixelscale, torch.tensor([[-0.173205, 0.1],[0.15,0.259808]], dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "WCS pixelscale should be set by state")
        self.assertTrue(torch.allclose(wcs.reference_radec, torch.tensor((120.2,-19), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "WCS reference RA DEC should be set by state")
        self.assertTrue(torch.allclose(wcs.reference_imagexy, torch.tensor((33., 123.), dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)), "WCS reference image position should be set by state")
        
    def test_wcs_repr(self):

        wcs = ap.image.WCS(
            projection = "orthographic",
            pixelscale = [[-0.173205, 0.1],[0.15,0.259808]],
            reference_radec = (120.2,-19),
            reference_imagexy = (33., 123.),
        )

        S = str(wcs)
        R = repr(wcs)



import numpy as np
import torch

from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from .superellipse_model import SuperEllipse_Galaxy, SuperEllipse_Warp
from .foureirellipse_model import FourierEllipse_Galaxy, FourierEllipse_Warp
from .psf_model_object import PSF_Model
from .ray_model import Ray_Galaxy
from .wedge_model import Wedge_Galaxy
from ..utils.decorators import ignore_numpy_warnings, default_internal

__all__ = [
    "RelSpline_Galaxy",
    "RelSpline_PSF",
]


# First Order
######################################################################
class RelSpline_Galaxy(Galaxy_Model):
    """Basic galaxy model with a spline radial light profile. The
    light profile is defined as a cubic spline interpolation of the
    stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I0: Central brightness
        dI(R): Tensor of brighntess values relative to central brightness, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"relspline {Galaxy_Model.model_type}"
    parameter_specs = {
        "I0": {"units": "log10(flux/arcsec^2)"},
        "dI(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("I0", "dI(R)")
    useable = True
    extend_profile = True

    from ._shared_methods import relspline_initialize as initialize
    from ._shared_methods import relspline_radial_model as radial_model


class RelSpline_PSF(PSF_Model):
    """point source model with a spline radial light profile. The light
    profile is defined as a cubic spline interpolation of the stored
    brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I0: Central brightness
        dI(R): Tensor of brighntess values relative to central brightness, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"relspline {PSF_Model.model_type}"
    parameter_specs = {
        "I0": {"units": "log10(flux/arcsec^2)", "value": 0., "locked": True},
        "dI(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = PSF_Model._parameter_order + ("I0", "dI(R)")
    useable = True
    extend_profile = True
    model_integrated = False
    
    @default_internal
    def transform_coordinates(self, X=None, Y=None, image=None, parameters=None):
        return X, Y

    from ._shared_methods import relspline_initialize as initialize
    from ._shared_methods import relspline_radial_model as radial_model
    from ._shared_methods import radial_evaluate_model as evaluate_model



import unittest
from astrophot.param import Node as BaseNode, Parameter_Node, Param_Mask, Param_Unlock, Param_SoftLimits
import astrophot as ap
import torch
import numpy as np

class Node(BaseNode):
    """
    Dummy class for testing purposes
    """

    def value(self):
        return None

class TestNode(unittest.TestCase):

    def test_node_init(self):
        node1 = Node("node1")
        node2 = Node("node2", locked = True)

        # Check for bad nameing
        with self.assertRaises(ValueError):
            node_bad = Node("node:bad")

    def test_node_link(self):
        node1 = Node("node1")
        node2 = Node("node2")
        node3 = Node("node3", locked = True)

        node1.link(node2, node3)

        self.assertTrue(node1.branch, "node1 is a branch")
        self.assertFalse(node3.branch, "node1 is not a branch")
        self.assertIs(node1["node2"], node2, "node getitem should fetch correct node")

        for Na, Nb in zip(node1.flat().values(), (node2, node3)):
            self.assertIs(Na, Nb, "node flat should produce correct order")

        node4 = Node("node4")

        node2.link(node4)

        for Na, Nb in zip(node1.flat(include_locked=False).values(), (node4,)):
            self.assertIs(Na, Nb, "node flat should produce correct order")

        # Check for cycle in DAG
        with self.assertRaises(ap.errors.InvalidParameter):
            node4.link(node1)

        node1.dump()

        self.assertEqual(len(node1.nodes), 0, "dump should clear all nodes")


    def test_node_access(self):
        node1 = Node("node1")
        node2 = Node("node2")
        node3 = Node("node3", locked = True)

        node1.link(node2, node3)
        node4 = Node("node4")

        node2.link(node4)
        
        self.assertIs(node1["node2:node4"], node4, "node getitem should fetch correct node")
        self.assertEqual(node1["node1"], node1, "node should get itself when getter called with its name")

        # Check that error is raised when requesting non existent node
        with self.assertRaises(KeyError):
            badnode = node1[1.2]
            
    def test_state(self):
        
        node1 = Node("node1")
        node2 = Node("node2")
        node3 = Node("node3", locked = True)

        node1.link(node2, node3)

        state = node1.get_state()

        S = str(node1)
        R = repr(node1)

class TestParameter(unittest.TestCase):
    @torch.no_grad()
    def test_parameter_setting(self):
        base_param = Parameter_Node("base param")
        base_param.value = 1.0
        self.assertEqual(base_param.value, 1, msg="Value should be set to 1")
        
        base_param.value = 2.0
        self.assertEqual(base_param.value, 2, msg="Value should update to 2")
        
        base_param.value += 2.0
        self.assertEqual(base_param.value, 4, msg="Value should update to 4")

        # Test a locked parameter that it does not change
        locked_param = Parameter_Node("locked param", value=1.0, locked=True)
        locked_param.value = 2.0
        self.assertEqual(
            locked_param.value, 1, msg="Locked value should remain at 1"
        )

        locked_param.value = 2.0
        self.assertEqual(
            locked_param.value, 1, msg="Locked value should remain at 1"
        )
        
    def test_parameter_limits(self):

        # Lower limit parameter
        lowlim_param = Parameter_Node("lowlim param", limits=(1, None))
        lowlim_param.value = 100.0
        self.assertEqual(
            lowlim_param.value,
            100,
            msg="lower limit variable should not have upper limit",
        )
        with self.assertRaises(ap.errors.InvalidParameter):
            lowlim_param.value = -100.0

        # Upper limit parameter
        uplim_param = Parameter_Node("uplim param", limits=(None, 1))
        uplim_param.value = -100.0
        self.assertEqual(
            uplim_param.value,
            -100,
            msg="upper limit variable should not have lower limit",
        )
        with self.assertRaises(ap.errors.InvalidParameter):
            uplim_param.value = 100.0

        # Range limit parameter
        range_param = Parameter_Node("range param", limits=(-1, 1))
        with self.assertRaises(ap.errors.InvalidParameter):
            range_param.value = 100.0
        with self.assertRaises(ap.errors.InvalidParameter):
            range_param.value = -100.0

        # Cyclic Range limit parameter
        cyrange_param = Parameter_Node("cyrange param", limits=(-1, 1), cyclic=True)
        cyrange_param.value = 2.0
        self.assertEqual(
            cyrange_param.value,
            0,
            msg="cyclic variable should loop in range (upper)",
        )
        cyrange_param.value = -2.0
        self.assertEqual(
            cyrange_param.value,
            0,
            msg="cyclic variable should loop in range (lower)",
        )

    def test_parameter_array(self):

        param_array1 = Parameter_Node("array1", value=list(float(3 + i) for i in range(5)))
        param_array2 = Parameter_Node("array2", value=list(float(i) for i in range(5)))

        param_array2.value = list(float(3) for i in range(5))
        self.assertTrue(
            torch.all(param_array2.value == 3),
            msg="parameter array value should be updated",
        )

        self.assertEqual(
            len(param_array2), 5, "parameter array should have length attribute"
        )

    def test_parameter_gradients(self):
        V = torch.ones(3)
        V.requires_grad = True
        params = Parameter_Node("input params", value=V)
        X = torch.sum(params.value * 3)
        X.backward()
        self.assertTrue(torch.all(V.grad == 3), "Parameters should track gradient")

    def test_parameter_state(self):

        P = Parameter_Node("state", value = 1., uncertainty = 0.5, limits = (-2, 2), locked = True, prof = 1.)

        P2 = Parameter_Node("v2")
        P2.set_state(P.get_state())

        self.assertEqual(P.value, P2.value, "state should preserve value")
        self.assertEqual(P.uncertainty, P2.uncertainty, "state should preserve uncertainty")
        self.assertEqual(P.prof, P2.prof, "state should preserve prof")
        self.assertEqual(P.locked, P2.locked, "state should preserve locked")
        self.assertEqual(P.limits[0].tolist(), P2.limits[0].tolist(), "state should preserve limits")
        self.assertEqual(P.limits[1].tolist(), P2.limits[1].tolist(), "state should preserve limits")

        S = str(P)

    def test_parameter_value(self):

        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.5, limits = (-1, 1), locked = False, prof = 1.)

        P2 = Parameter_Node("test2", value = P1)

        P3 = Parameter_Node("test3", value = lambda P: P["test1"].value**2, link = (P1,))

        self.assertEqual(P1.value.item(), 0.5, "Parameter should store value")
        self.assertEqual(P2.value.item(), 0.5, "Pointing parameter should fetch value")
        self.assertEqual(P3.value.item(), 0.25, "Function parameter should compute value")

        self.assertEqual(P2.shape, P1.shape, "reference node should map shape")
        self.assertEqual(P3.shape, P1.shape, "reference node should map shape")
        
class TestParamContext(unittest.TestCase):
    def test_unlock(self):
        locked_param = Parameter_Node("locked param", value=1.0, locked=True)
        locked_param.value = 2.
        self.assertEqual(locked_param.value.item(), 1., "locked parameter should not be updated out of context")
        with Param_Unlock(locked_param):
            locked_param.value = 2.
        self.assertEqual(locked_param.value.item(), 2., "locked parameter should be updated in context")
        with Param_Unlock():
            locked_param.value = 3.
        self.assertEqual(locked_param.value.item(), 3., "locked parameter should be updated in global unlock context")
        
            
class TestParameterVector(unittest.TestCase):
    def test_param_vector_creation(self):

        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.5, limits = (-1, 1), locked = False, prof = 1.)
        P2 = Parameter_Node("test2", value = 2., uncertainty = 5., locked = False)
        P3 = Parameter_Node("test3", value = [4.,5.], uncertainty = [5.,5.], locked = False)
        P4 = Parameter_Node("test4", value = P2)
        P5 = Parameter_Node("test5", value = lambda P: P["test1"].value**2, link = (P1,))
        PG = Parameter_Node("testgroup", link = (P1, P2, P3, P4, P5))
        
        self.assertTrue(torch.all(PG.vector_values() == torch.tensor([0.5,2.,4.,5.], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node values")
        self.assertEqual(PG.mask.numel(), 4, "Vector should take all/only leaf node masks")
        self.assertEqual(PG.vector_identities().size, 4, "Vector should take all/only leaf node identities")
        self.assertEqual(PG.identities.size, 4, "Vector should take all/only leaf node identities")
        self.assertEqual(PG.names.size, 4, "Vector should take all/only leaf node names")
        self.assertEqual(PG.vector_names().size, 4, "Vector should take all/only leaf node names")

        PG.value = [1.,2.,3.,4.]
        self.assertTrue(torch.all(PG.vector_values() == torch.tensor([1.,2.,3.,4.], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node values")
        
    def test_vector_masking(self):
        
        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.3, limits = (-1, 1), locked = False, prof = 1.)
        P2 = Parameter_Node("test2", value = 2., uncertainty = 1., locked = False)
        P3 = Parameter_Node("test3", value = [4.,5.], uncertainty = [5.,3.], locked = False)
        P4 = Parameter_Node("test4", value = P2)
        P5 = Parameter_Node("test5", value = lambda P: P["test1"].value**2, link = (P1,))
        PG = Parameter_Node("testgroup", link = (P1, P2, P3, P4, P5))

        mask = torch.tensor([1,0,0,1], dtype = torch.bool, device=P1.value.device)

        with Param_Mask(PG, mask):
            self.assertTrue(torch.all(PG.vector_values() == torch.tensor([0.5,5.], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node values")
            self.assertTrue(torch.all(PG.vector_uncertainty() == torch.tensor([0.3,3.], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node uncertainty")
            self.assertEqual(PG.vector_mask().numel(), 4, "Vector should take all/only leaf node masks")
            self.assertEqual(PG.vector_identities().size, 2, "Vector should take all/only leaf node identities")

            # Nested masking
            new_mask = torch.tensor([1,0], dtype = torch.bool, device=P1.value.device)
            with Param_Mask(PG, new_mask):
                self.assertTrue(torch.all(PG.vector_values() == torch.tensor([0.5], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node values")
                self.assertTrue(torch.all(PG.vector_uncertainty() == torch.tensor([0.3], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node uncertainty")
                self.assertEqual(PG.vector_mask().numel(), 4, "Vector should take all/only leaf node masks")
                self.assertEqual(PG.vector_identities().size, 1, "Vector should take all/only leaf node identities")

            self.assertTrue(torch.all(PG.vector_values() == torch.tensor([0.5,5.], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node values")
            self.assertTrue(torch.all(PG.vector_uncertainty() == torch.tensor([0.3,3.], dtype=P1.value.dtype, device = P1.value.device)), "Vector store all leaf node uncertainty")
            self.assertEqual(PG.vector_mask().numel(), 4, "Vector should take all/only leaf node masks")
            self.assertEqual(PG.vector_identities().size, 2, "Vector should take all/only leaf node identities")


    def test_vector_representation(self):
        
        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.3, limits = (-1, 1), locked = False, prof = 1.)
        P2 = Parameter_Node("test2", value = 2., uncertainty = 1., locked = False)
        P3 = Parameter_Node("test3", value = [4.,5.], uncertainty = [5.,3.], limits = ((0., 1.), None), locked = False)
        P4 = Parameter_Node("test4", value = P2)
        P5 = Parameter_Node("test5", value = lambda P: P["test1"].value**2, link = (P1,))
        P6 = Parameter_Node("test6", value = ((5,6),(7,8)), uncertainty = 0.1 * np.zeros((2,2)), limits = (None, 10*np.ones((2,2))))
        PG = Parameter_Node("testgroup", link = (P1, P2, P3, P4, P5, P6))

        mask = torch.tensor([1,1,0,1,0,1,0,1], dtype = torch.bool, device=P1.value.device)

        self.assertEqual(len(PG.vector_representation()), 8, "representation should collect all values")
        with Param_Mask(PG, mask):
            # round trip
            vec = PG.vector_values().clone()
            rep = PG.vector_representation()
            PG.vector_set_representation(rep)
            self.assertTrue(torch.all(vec == PG.vector_values()), "representation should be reversible")
            self.assertEqual(PG.vector_values().numel(), 5, "masked values shouldn't be shown")


    def test_printing(self):

        def node_func_sqr(P):
            return P["test1"].value**2
        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.3, limits = (-1, 1), locked = False, prof = 1.)
        P2 = Parameter_Node("test2", value = 2., uncertainty = 1., locked = False)
        P3 = Parameter_Node("test3", value = [4.,5.], uncertainty = [5.,3.], limits = ((0., 1.), None), locked = False)
        P4 = Parameter_Node("test4", value = P2)
        P5 = Parameter_Node("test5", value = node_func_sqr, link = (P1,))
        P6 = Parameter_Node("test6", value = ((5,6),(7,8)), uncertainty = 0.1 * np.zeros((2,2)), limits = (None, 10*np.ones((2,2))))
        PG = Parameter_Node("testgroup", link = (P1, P2, P3, P4, P5, P6))

        self.assertEqual(str(PG), """testgroup:
test1: 0.5 +- 0.3 [none], limits: (-1.0, 1.0)
test2: 2.0 +- 1.0 [none]
test3: [4.0, 5.0] +- [5.0, 3.0] [none], limits: ([0.0, 1.0], None)
test6: [[5.0, 6.0], [7.0, 8.0]] +- [[0.0, 0.0], [0.0, 0.0]] [none], limits: (None, [[10.0, 10.0], [10.0, 10.0]])""", "String representation should return specific string")

        ref_string = """testgroup (id-140071931416000, branch node):
  test1 (id-140071931414752): 0.5 +- 0.3 [none], limits: (-1.0, 1.0)
  test2 (id-140071931415376): 2.0 +- 1.0 [none]
  test3 (id-140071931415472): [4.0, 5.0] +- [5.0, 3.0] [none], limits: ([0.0, 1.0], None)
  test4 (id-140071931414272) points to: test2 (id-140071931415376): 2.0 +- 1.0 [none]
  test5 (id-140071931414992, function node, node_func_sqr):
    test1 (id-140071931414752): 0.5 +- 0.3 [none], limits: (-1.0, 1.0)
  test6 (id-140071931415616): [[5.0, 6.0], [7.0, 8.0]] +- [[0.0, 0.0], [0.0, 0.0]] [none], limits: (None, [[10.0, 10.0], [10.0, 10.0]])"""
        # Remove ids since they change every time
        while "(id-" in ref_string:
            start = ref_string.find("(id-")
            end = ref_string.find(")", start)+1
            ref_string = ref_string[:start] + ref_string[end:]

        repr_string = repr(PG)
        # Remove ids since they change every time
        count = 0
        while "(id-" in repr_string:
            start = repr_string.find("(id-")
            end = repr_string.find(")", start)+1
            repr_string = repr_string[:start] + repr_string[end:]
            count += 1
            if count > 100:
                raise RuntimeError("infinite loop! Something very wrong with parameter repr")
        self.assertEqual(repr_string, ref_string, "Repr should return specific string")


    def test_empty_vector(self):
        def node_func_sqr(P):
            return P["test1"].value**2
        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.3, limits = (-1, 1), locked = True, prof = 1.)
        P2 = Parameter_Node("test2", value = 2., uncertainty = 1., locked = True)
        P3 = Parameter_Node("test3", value = [4.,5.], uncertainty = [5.,3.], limits = ((0., 1.), None), locked = True)
        P4 = Parameter_Node("test4", value = P2)
        P5 = Parameter_Node("test5", value = node_func_sqr, link = (P1,))
        P6 = Parameter_Node("test6", value = ((5,6),(7,8)), uncertainty = 0.1 * np.zeros((2,2)), limits = (None, 10*np.ones((2,2))), locked = True)
        PG = Parameter_Node("testgroup", link = (P1, P2, P3, P4, P5, P6))

        self.assertEqual(PG.names.shape, (0,), "all locked parameter should have empty names")
        self.assertEqual(PG.identities.shape,(0,), "all locked parameter should have empty identities")
        self.assertEqual(PG.vector_names().shape, (0,), "all locked parameter should have empty names")
        self.assertEqual(PG.vector_identities().shape,(0,), "all locked parameter should have empty identities")

        self.assertEqual(PG.vector_values().shape, (0,), "all locked parameter should have empty values")
        self.assertEqual(PG.vector_uncertainty().shape, (0,), "all locked parameter should have empty uncertainty")
        self.assertEqual(PG.vector_mask().shape, (0,), "all locked parameter should have empty mask")
        self.assertEqual(PG.vector_representation().shape, (0,), "all locked parameter should have empty representation")

    def test_none_uncertainty(self):
        
        P1 = Parameter_Node("test1", value = 0.5, uncertainty = 0.3, limits = (-1, 1), locked = False, prof = 1.)
        P2 = Parameter_Node("test2", value = 2., locked = True)
        P3 = Parameter_Node("test3", value = [4.,5.], limits = ((0., 1.), None), locked = False)
        P4 = Parameter_Node("test4", link = (P1, P2, P3))

        self.assertEqual(tuple(P4.vector_uncertainty().detach().cpu().tolist()), (0.3, 1., 1.), "None uncertainty should be filled with ones")
        
        P3.uncertainty = None
        P4.vector_set_uncertainty((0.1,0.1,0.1))
        
        self.assertEqual(tuple(P4.vector_uncertainty().detach().cpu().tolist()), (0.1, 0.1, 0.1), "None uncertainty should be filled using vector_set_uncertainty")
        
if __name__ == "__main__":
    unittest.main()



import unittest
import numpy as np
import torch
import h5py
from scipy.signal import fftconvolve
from scipy.special import gamma
from torch.special import gammaln
from scipy.interpolate import RectBivariateSpline
import astrophot as ap
from utils import make_basic_sersic, make_basic_gaussian

######################################################################
# Util functions
######################################################################


class TestFFT(unittest.TestCase):
    def test_fft(self):

        target = make_basic_sersic()

        convolved = ap.utils.operations.fft_convolve_torch(
            target.data,
            target.psf.data,
        )
        scipy_convolve = fftconvolve(
            target.data.detach().cpu().numpy(),
            target.psf.data.detach().cpu().numpy(),
            mode="same",
        )
        self.assertLess(
            torch.std(convolved),
            torch.std(target.data),
            "Convolved image should be smoothed",
        )

        self.assertTrue(
            np.all(np.isclose(convolved.detach().cpu().numpy(), scipy_convolve)),
            "Should reproduce scipy convolve",
        )

    def test_fft_multi(self):

        target = make_basic_sersic()

        convolved = ap.utils.operations.fft_convolve_multi_torch(
            target.data, [target.psf.data, target.psf.data]
        )
        self.assertLess(
            torch.std(convolved),
            torch.std(target.data),
            "Convolved image should be smoothed",
        )


class TestOptimize(unittest.TestCase):
    def test_chi2(self):

        # with variance
        # with mask
        mask = torch.zeros(10, dtype=torch.bool, device = ap.AP_config.ap_device)
        mask[2] = 1
        chi2 = ap.utils.optimization.chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            mask=mask, variance=2 * torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)
        )
        self.assertEqual(chi2, 4.5, "Chi squared calculation incorrect")
        chi2_red = ap.utils.optimization.reduced_chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            params=3,
            mask=mask,
            variance=2 * torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
        )
        self.assertEqual(chi2_red.item(), 0.75, "Chi squared calculation incorrect")

        # no mask
        chi2 = ap.utils.optimization.chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            variance=2 * torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
        )
        self.assertEqual(chi2, 5, "Chi squared calculation incorrect")
        chi2_red = ap.utils.optimization.reduced_chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            params=3, variance=2 * torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)
        )
        self.assertEqual(chi2_red.item(), 5 / 7, "Chi squared calculation incorrect")

        # no variance
        # with mask
        mask = torch.zeros(10, dtype=torch.bool, device = ap.AP_config.ap_device)
        mask[2] = 1
        chi2 = ap.utils.optimization.chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            mask=mask
        )
        self.assertEqual(chi2.item(), 9, "Chi squared calculation incorrect")
        chi2_red = ap.utils.optimization.reduced_chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            params=3, mask=mask
        )
        self.assertEqual(chi2_red.item(), 1.5, "Chi squared calculation incorrect")

        # no mask
        chi2 = ap.utils.optimization.chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device)
        )
        self.assertEqual(chi2.item(), 10, "Chi squared calculation incorrect")
        chi2_red = ap.utils.optimization.reduced_chi_squared(
            torch.ones(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            torch.zeros(10, dtype = ap.AP_config.ap_dtype, device = ap.AP_config.ap_device),
            params=3
        )
        self.assertEqual(chi2_red.item(), 10 / 7, "Chi squared calculation incorrect")


class TestPSF(unittest.TestCase):
    def test_make_psf(self):

        target = make_basic_gaussian(x=10, y=10)
        target += make_basic_gaussian(x=40, y=40, rand=54321)

        psf = ap.utils.initialize.construct_psf(
            [[10, 10], [40, 40]],
            target.data.detach().cpu().numpy(),
            sky_est=0.0,
            size=5,
        )

        self.assertTrue(np.all(np.isfinite(psf)))


class TestSegtoWindow(unittest.TestCase):
    def test_segtowindow(self):

        segmap = np.zeros((100, 100), dtype=int)

        segmap[5:9, 20:30] = 1
        segmap[50:90, 17:35] = 2
        segmap[26:34, 80:85] = 3

        centroids = ap.utils.initialize.centroids_from_segmentation_map(
            segmap, image=segmap
        )

        PAs = ap.utils.initialize.PA_from_segmentation_map(
            segmap, image=segmap, centroids = centroids,
        )
        qs = ap.utils.initialize.q_from_segmentation_map(
            segmap, image=segmap, centroids = centroids,
        )

        windows = ap.utils.initialize.windows_from_segmentation_map(segmap)

        self.assertEqual(
            len(windows), 3, "should ignore zero index, but find all three windows"
        )
        self.assertEqual(
            len(centroids), 3, "should ignore zero index, but find all three windows"
        )
        self.assertEqual(
            len(PAs), 3, "should ignore zero index, but find all three windows"
        )
        self.assertEqual(
            len(qs), 3, "should ignore zero index, but find all three windows"
        )

        self.assertEqual(
            windows[1], [[20, 29], [5, 8]], "Windows should be identified by index"
        )

        # scale windows

        new_windows = ap.utils.initialize.scale_windows(
            windows, image_shape=(100, 100), expand_scale=2, expand_border=3
        )

        self.assertEqual(
            new_windows[2], [[5, 45], [27, 100]], "Windows should scale appropriately"
        )

        filtered_windows = ap.utils.initialize.filter_windows(
            new_windows, min_size=10, max_size=80, min_area=30, max_area=1000
        )
        filtered_windows = ap.utils.initialize.filter_windows(
            new_windows, min_flux=10, max_flux=1000, image=np.ones(segmap.shape)
        )

        self.assertEqual(len(filtered_windows), 2, "windows should have been filtered")

        # check original
        self.assertEqual(
            windows[3], [[80, 84], [26, 33]], "Original windows should not have changed"
        )


class TestConversions(unittest.TestCase):
    def test_conversions_units(self):

        # flux to sb
        self.assertEqual(
            ap.utils.conversions.units.flux_to_sb(1.0, 1.0, 0.0),
            0,
            "flux incorrectly converted to sb",
        )

        # sb to flux
        self.assertEqual(
            ap.utils.conversions.units.sb_to_flux(1.0, 1.0, 0.0),
            (10 ** (-1 / 2.5)),
            "sb incorrectly converted to flux",
        )

        # flux to mag no error
        self.assertEqual(
            ap.utils.conversions.units.flux_to_mag(1.0, 0.0),
            0,
            "flux incorrectly converted to mag (no error)",
        )

        # flux to mag with error
        self.assertEqual(
            ap.utils.conversions.units.flux_to_mag(1.0, 0.0, fluxe=1.0),
            (0.0, 2.5 / np.log(10)),
            "flux incorrectly converted to mag (with error)",
        )

        # mag to flux no error:
        self.assertEqual(
            ap.utils.conversions.units.mag_to_flux(1.0, 0.0, mage=None),
            (10 ** (-1 / 2.5)),
            "mag incorrectly converted to flux (no error)",
        )

        # mag to flux with error:
        [
            self.assertAlmostEqual(
                ap.utils.conversions.units.mag_to_flux(1.0, 0.0, mage=1.0)[i],
                (10 ** (-1.0 / 2.5), np.log(10) * (1.0 / 2.5) * 10 ** (-1.0 / 2.5))[i],
                msg="mag incorrectly converted to flux (with error)",
            )
            for i in range(1)
        ]

        # magperarcsec2 to mag with area A defined
        self.assertAlmostEqual(
            ap.utils.conversions.units.magperarcsec2_to_mag(1.0, a=None, b=None, A=1.0),
            (1.0 - 2.5 * np.log10(1.0)),
            msg="mag/arcsec^2 incorrectly converted to mag (area A given, a and b not defined)",
        )

        # magperarcsec2 to mag with semi major and minor axes defined (a, and b)
        self.assertAlmostEqual(
            ap.utils.conversions.units.magperarcsec2_to_mag(1.0, a=1.0, b=1.0, A=None),
            (1.0 - 2.5 * np.log10(np.pi)),
            msg="mag/arcsec^2 incorrectly converted to mag (semi major/minor axes defined)",
        )

        # mag to magperarcsec2 with area A defined
        self.assertAlmostEqual(
            ap.utils.conversions.units.mag_to_magperarcsec2(
                1.0, a=None, b=None, A=1.0, R=None
            ),
            (1.0 + 2.5 * np.log10(1.0)),
            msg="mag incorrectly converted to mag/arcsec^2 (area A given)",
        )

        # mag to magperarcsec2 with radius R given (assumes circular)
        self.assertAlmostEqual(
            ap.utils.conversions.units.mag_to_magperarcsec2(
                1.0, a=None, b=None, A=None, R=1.0
            ),
            (1.0 + 2.5 * np.log10(np.pi)),
            msg="mag incorrectly converted to mag/arcsec^2 (radius R given)",
        )

        # mag to magperarcsec2 with semi major and minor axes defined (a, and b)
        self.assertAlmostEqual(
            ap.utils.conversions.units.mag_to_magperarcsec2(
                1.0, a=1.0, b=1.0, A=None, R=None
            ),
            (1.0 + 2.5 * np.log10(np.pi)),
            msg="mag incorrectly converted to mag/arcsec^2 (area A given)",
        )

        # position angle PA to radians
        self.assertAlmostEqual(
            ap.utils.conversions.units.PA_shift_convention(1.0, unit="rad"),
            ((1.0 - (np.pi / 2)) % np.pi),
            msg="PA incorrectly converted to radians",
        )

        # position angle PA to degrees
        self.assertAlmostEqual(
            ap.utils.conversions.units.PA_shift_convention(1.0, unit="deg"),
            ((1.0 - (180 / 2)) % 180),
            msg="PA incorrectly converted to degrees",
        )

    def test_conversion_dict_to_hdf5(self):

        # convert string to hdf5
        self.assertEqual(
            ap.utils.conversions.dict_to_hdf5.to_hdf5_has_None(l="test"),
            (False),
            "Failed to properly identify string object while converting to hdf5",
        )

        # convert __iter__ to hdf5
        self.assertEqual(
            ap.utils.conversions.dict_to_hdf5.to_hdf5_has_None(l="__iter__"),
            (False),
            "Attempted to convert '__iter__' to hdf5 key",
        )

        # convert hdf5 file to dict
        h = h5py.File("mytestfile.hdf5", "w")
        dset = h.create_dataset("mydataset", (1,), dtype="i")
        dset[...] = np.array([1.0])
        self.assertEqual(
            ap.utils.conversions.dict_to_hdf5.hdf5_to_dict(h=h),
            ({"mydataset": h["mydataset"]}),
            "Failed to convert hdf5 file to dict",
        )

        # convert dict to hdf5
        target = make_basic_sersic().data.detach().cpu().numpy()[0]
        d = {"sersic": target.tolist()}
        ap.utils.conversions.dict_to_hdf5.dict_to_hdf5(
            h=h5py.File("mytestfile2.hdf5", "w"), D=d
        )
        self.assertEqual(
            (list(h5py.File("mytestfile2.hdf5", "r"))),
            (list(d)),
            "Failed to convert dict of strings to hdf5",
        )

    def test_conversion_functions(self):

        sersic_n = ap.utils.conversions.functions.sersic_n_to_b(1.0)
        # sersic I0 to flux - numpy
        self.assertAlmostEqual(
            ap.utils.conversions.functions.sersic_I0_to_flux_np(1.0, 1.0, 1.0, 1.0),
            (2 * np.pi * gamma(2)),
            msg="Error converting sersic central intensity to flux (np)",
        )

        # sersic flux to I0 - numpy
        self.assertAlmostEqual(
            ap.utils.conversions.functions.sersic_flux_to_I0_np(1.0, 1.0, 1.0, 1.0),
            (1.0 / (2 * np.pi * gamma(2))),
            msg="Error converting sersic flux to central intensity (np)",
        )

        # sersic Ie to flux - numpy
        self.assertAlmostEqual(
            ap.utils.conversions.functions.sersic_Ie_to_flux_np(1.0, 1.0, 1.0, 1.0),
            (2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2)),
            msg="Error converting sersic effective intensity to flux (np)",
        )

        # sersic flux to Ie - numpy
        self.assertAlmostEqual(
            ap.utils.conversions.functions.sersic_flux_to_Ie_np(1.0, 1.0, 1.0, 1.0),
            (1 / (2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2))),
            msg="Error converting sersic flux to effective intensity (np)",
        )

        # inverse sersic - numpy
        self.assertAlmostEqual(
            ap.utils.conversions.functions.sersic_inv_np(1.0, 1.0, 1.0, 1.0),
            (1.0 - (1.0 / sersic_n) * np.log(1.0)),
            msg="Error computing inverse sersic function (np)",
        )

        # sersic I0 to flux - torch
        tv = torch.tensor([[1.0]], dtype=torch.float64)
        self.assertEqual(
            torch.round(
                ap.utils.conversions.functions.sersic_I0_to_flux_np(tv, tv, tv, tv),
                decimals=7,
            ),
            torch.round(torch.tensor([[2 * np.pi * gamma(2)]]), decimals=7),
            msg="Error converting sersic central intensity to flux (torch)",
        )

        # sersic flux to I0 - torch
        self.assertEqual(
            torch.round(
                ap.utils.conversions.functions.sersic_flux_to_I0_np(tv, tv, tv, tv),
                decimals=7,
            ),
            torch.round(torch.tensor([[1.0 / (2 * np.pi * gamma(2))]]), decimals=7),
            msg="Error converting sersic flux to central intensity (torch)",
        )

        # sersic Ie to flux - torch
        self.assertEqual(
            torch.round(
                ap.utils.conversions.functions.sersic_Ie_to_flux_np(tv, tv, tv, tv),
                decimals=7,
            ),
            torch.round(
                torch.tensor(
                    [[2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2)]]
                ),
                decimals=7,
            ),
            msg="Error converting sersic effective intensity to flux (torch)",
        )

        # sersic flux to Ie - torch
        self.assertEqual(
            torch.round(
                ap.utils.conversions.functions.sersic_flux_to_Ie_np(tv, tv, tv, tv),
                decimals=7,
            ),
            torch.round(
                torch.tensor(
                    [[1 / (2 * np.pi * gamma(2) * np.exp(sersic_n) * sersic_n ** (-2))]]
                ),
                decimals=7,
            ),
            msg="Error converting sersic flux to effective intensity (torch)",
        )

        # inverse sersic - torch
        self.assertEqual(
            torch.round(
                ap.utils.conversions.functions.sersic_inv_np(tv, tv, tv, tv), decimals=7
            ),
            torch.round(
                torch.tensor([[1.0 - (1.0 / sersic_n) * np.log(1.0)]]), decimals=7
            ),
            msg="Error computing inverse sersic function (torch)",
        )
    def test_general_derivative(self):

        res = ap.utils.conversions.functions.general_uncertainty_prop(
            tuple(torch.tensor(a) for a in (1.0, 1.0, 1.0, 0.5)),
            tuple(torch.tensor(a) for a in (0.1, 0.1, 0.1, 0.1)),
            ap.utils.conversions.functions.sersic_Ie_to_flux_torch,
        )

        self.assertAlmostEqual(res.detach().cpu().numpy(), 1.8105, 3, "General uncertianty prop should compute uncertainty")


class TestInterpolate(unittest.TestCase):
    def test_interpolate_functions(self):

        # Lanczos kernel interpolation on the center point of a gaussian (10., 10.)
        model = make_basic_gaussian(x=10.0, y=10.0).data.detach().cpu().numpy()
        lanczos_interp = ap.utils.interpolate.point_Lanczos(
            model, 10.0, 10.0, scale=0.8
        )
        self.assertTrue(
            np.all(np.isfinite(model)), msg="gaussian model returning nonfinite values"
        )
        self.assertLess(
            lanczos_interp, 1.0, msg="Lanczos interpolation greater than total flux"
        )
        self.assertTrue(
            np.isfinite(lanczos_interp),
            msg="Lanczos interpolate returning nonfinite values",
        )


class TestAngleOperations(unittest.TestCase):
    def test_angle_operation_functions(self):

        test_angles = np.array([np.pi, 2 * np.pi, 3 * np.pi, 4 * np.pi])
        # angle median
        self.assertAlmostEqual(
            ap.utils.angle_operations.Angle_Median(test_angles),
            -np.pi / 2,
            msg="incorrectly calculating median of list of angles",
        )

        # angle scatter (iqr)
        self.assertAlmostEqual(
            ap.utils.angle_operations.Angle_Scatter(test_angles),
            np.pi,
            msg="incorrectly calculating iqr of list of angles",
        )

    def test_angle_com(self):
        pixelscale = 0.8
        tar = make_basic_sersic(
            N=50,
            M=50,
            pixelscale=pixelscale,
            x=24.5*pixelscale,
            y=24.5*pixelscale,
            PA = 115 * np.pi / 180,
        )

        res = ap.utils.angle_operations.Angle_COM_PA(tar.data.detach().cpu().numpy())

        self.assertAlmostEqual(res + np.pi/2, 115 * np.pi / 180, delta = 0.1)
        
        
if __name__ == "__main__":
    unittest.main()



import numpy as np
import torch

from .galaxy_model_object import Galaxy_Model
from .warp_model import Warp_Galaxy
from .superellipse_model import SuperEllipse_Galaxy, SuperEllipse_Warp
from .foureirellipse_model import FourierEllipse_Galaxy, FourierEllipse_Warp
from .psf_model_object import PSF_Model
from .ray_model import Ray_Galaxy
from .wedge_model import Wedge_Galaxy
from ._shared_methods import spline_segment_initialize, select_target
from ..utils.decorators import ignore_numpy_warnings, default_internal

__all__ = [
    "Spline_Galaxy",
    "Spline_PSF",
    "Spline_Warp",
    "Spline_SuperEllipse",
    "Spline_FourierEllipse",
    "Spline_Ray",
    "Spline_SuperEllipse_Warp",
    "Spline_FourierEllipse_Warp",
]


# First Order
######################################################################
class Spline_Galaxy(Galaxy_Model):
    """Basic galaxy model with a spline radial light profile. The
    light profile is defined as a cubic spline interpolation of the
    stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {Galaxy_Model.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Galaxy_Model._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model


class Spline_PSF(PSF_Model):
    """star model with a spline radial light profile. The light
    profile is defined as a cubic spline interpolation of the stored
    brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {PSF_Model.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = PSF_Model._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True
    model_integrated = False
    
    @default_internal
    def transform_coordinates(self, X=None, Y=None, image=None, parameters=None):
        return X, Y

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model
    from ._shared_methods import radial_evaluate_model as evaluate_model


class Spline_Warp(Warp_Galaxy):
    """warped coordinate galaxy model with a spline light
    profile. The light profile is defined as a cubic spline
    interpolation of the stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {Warp_Galaxy.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Warp_Galaxy._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model


# Second Order
######################################################################
class Spline_SuperEllipse(SuperEllipse_Galaxy):
    """The light profile is defined as a cubic spline interpolation of
    the stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {SuperEllipse_Galaxy.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = SuperEllipse_Galaxy._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model


class Spline_FourierEllipse(FourierEllipse_Galaxy):
    """The light profile is defined as a cubic spline interpolation of the
    stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {FourierEllipse_Galaxy.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = FourierEllipse_Galaxy._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model


class Spline_Ray(Ray_Galaxy):
    """ray galaxy model with a spline light profile. The light
    profile is defined as a cubic spline interpolation of the stored
    brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): 2D Tensor of brighntess values for each ray, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {Ray_Galaxy.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Ray_Galaxy._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        spline_segment_initialize(
            self,
            target=target,
            parameters=parameters,
            segments=self.rays,
            symmetric=self.symmetric_rays,
        )

    from ._shared_methods import spline_iradial_model as iradial_model


class Spline_Wedge(Wedge_Galaxy):
    """wedge galaxy model with a spline light profile. The light
    profile is defined as a cubic spline interpolation of the stored
    brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): 2D Tensor of brighntess values for each wedge, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {Wedge_Galaxy.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = Wedge_Galaxy._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    @torch.no_grad()
    @ignore_numpy_warnings
    @select_target
    @default_internal
    def initialize(self, target=None, parameters=None, **kwargs):
        super().initialize(target=target, parameters=parameters)

        spline_segment_initialize(
            self,
            target=target,
            parameters=parameters,
            segments=self.wedges,
            symmetric=self.symmetric_wedges,
        )

    from ._shared_methods import spline_iradial_model as iradial_model


# Third Order
######################################################################
class Spline_SuperEllipse_Warp(SuperEllipse_Warp):
    """The light profile is defined as a cubic spline interpolation of the
    stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {SuperEllipse_Warp.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = SuperEllipse_Warp._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model


class Spline_FourierEllipse_Warp(FourierEllipse_Warp):
    """The light profile is defined as a cubic spline interpolation of the
    stored brightness values:

    I(R) = interp(R, profR, I)

    where I(R) is the brightness along the semi-major axis, interp is
    a cubic spline function, R is the semi-major axis length, profR is
    a list of radii for the spline, I is a corresponding list of
    brightnesses at each profR value.

    Parameters:
        I(R): Tensor of brighntess values, represented as the log of the brightness divided by pixelscale squared

    """

    model_type = f"spline {FourierEllipse_Warp.model_type}"
    parameter_specs = {
        "I(R)": {"units": "log10(flux/arcsec^2)"},
    }
    _parameter_order = FourierEllipse_Warp._parameter_order + ("I(R)",)
    useable = True
    extend_profile = True

    from ._shared_methods import spline_initialize as initialize
    from ._shared_methods import spline_radial_model as radial_model



import unittest
import astrophot as ap
import numpy as np
import torch

######################################################################
# Image List Object
######################################################################


class TestImageList(unittest.TestCase):
    def test_image_creation(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
            metadata = {"note":"test image 1"},
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
            metadata = {"note":"test image 2"},
        )

        test_image = ap.image.Image_List((base_image1, base_image2))

        for image, original_image in zip(test_image, (base_image1, base_image2)):
            self.assertEqual(
                image.pixel_length,
                original_image.pixel_length,
                "image should track pixelscale",
            )
            self.assertEqual(
                image.zeropoint,
                original_image.zeropoint,
                "image should track zeropoint",
            )
            self.assertEqual(
                image.origin[0], original_image.origin[0], "image should track origin"
            )
            self.assertEqual(
                image.origin[1], original_image.origin[1], "image should track origin"
            )
            self.assertEqual(image.metadata["note"], original_image.metadata["note"], "image should track note")

        slicer = ap.image.Window_List(
            (ap.image.Window(origin=(3, 2), pixel_shape=(4, 5)), ap.image.Window(origin=(3, 2), pixel_shape=(4, 5)))
        )
        sliced_image = test_image[slicer]

        self.assertEqual(sliced_image[0].origin[0], 3, "image should track origin")
        self.assertEqual(sliced_image[0].origin[1], 2, "image should track origin")
        self.assertEqual(sliced_image[1].origin[0], 3, "image should track origin")
        self.assertEqual(sliced_image[1].origin[1], 2, "image should track origin")
        self.assertEqual(
            base_image1.origin[0], 0, "subimage should not change image origin"
        )
        self.assertEqual(
            base_image1.origin[1], 0, "subimage should not change image origin"
        )

    def test_copy(self):

        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )

        test_image = ap.image.Image_List((base_image1, base_image2))

        copy_image = test_image.copy()
        for ti, ci in zip(test_image, copy_image):
            self.assertEqual(
                ti.pixel_length, ci.pixel_length, "copied image should have same pixelscale"
            )
            self.assertEqual(
                ti.zeropoint, ci.zeropoint, "copied image should have same zeropoint"
            )
            self.assertEqual(
                ti.window, ci.window, "copied image should have same window"
            )
            preval = ti.data[0][0].item()
            ci += 1
            self.assertEqual(
                ti.data[0][0],
                preval,
                "copied image should not share data with original",
            )

        blank_copy_image = test_image.blank_copy()
        for ti, ci in zip(test_image, blank_copy_image):
            self.assertEqual(
                ti.pixel_length, ci.pixel_length, "copied image should have same pixelscale"
            )
            self.assertEqual(
                ti.zeropoint, ci.zeropoint, "copied image should have same zeropoint"
            )
            self.assertEqual(
                ti.window, ci.window, "copied image should have same window"
            )
            preval = ti.data[0][0].item()
            ci += 1
            self.assertEqual(
                ti.data[0][0],
                preval,
                "copied image should not share data with original",
            )

    def test_image_arithmetic(self):

        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )
        test_image = ap.image.Image_List((base_image1, base_image2))

        arr3 = torch.ones((10, 15))
        base_image3 = ap.image.Image(
            data=arr3,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.ones(2),
        )
        arr4 = torch.zeros((15, 10))
        base_image4 = ap.image.Image(
            data=arr4,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.zeros(2),
        )
        second_image = ap.image.Image_List((base_image3, base_image4))

        # Test iadd
        test_image += second_image

        self.assertEqual(
            test_image[0].data[0][0], 0, "image addition should only update its region"
        )
        self.assertEqual(
            test_image[0].data[3][3], 1, "image addition should update its region"
        )
        self.assertEqual(
            test_image[1].data[0][0], 1, "image addition should update its region"
        )
        self.assertEqual(
            test_image[1].data[1][1], 1, "image addition should update its region"
        )

        # Test iadd
        test_image -= second_image

        self.assertEqual(
            test_image[0].data[0][0], 0, "image addition should only update its region"
        )
        self.assertEqual(
            test_image[0].data[3][3], 0, "image addition should update its region"
        )
        self.assertEqual(
            test_image[1].data[0][0], 1, "image addition should update its region"
        )
        self.assertEqual(
            test_image[1].data[1][1], 1, "image addition should update its region"
        )

    def test_image_list_display(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )
        test_image = ap.image.Image_List((base_image1, base_image2))

        self.assertIsInstance(str(test_image), str, "String representation should be a string!")
        self.assertIsInstance(repr(test_image), str, "Repr should be a string!")

    def test_image_list_windowset(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
            note="test image 1",
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
            note="test image 2",
        )
        test_image = ap.image.Image_List((base_image1, base_image2))
        arr3 = torch.ones((10, 15))
        base_image3 = ap.image.Image(
            data=arr3,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.ones(2),
            note="test image 3",
        )
        arr4 = torch.zeros((15, 10))
        base_image4 = ap.image.Image(
            data=arr4,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.zeros(2),
            note="test image 4",
        )
        second_image = ap.image.Image_List((base_image3, base_image4), window = test_image.window)       

    def test_image_list_errors(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )
        test_image = ap.image.Image_List((base_image1, base_image2))
        # Bad ra dec reference point
        bad_base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            reference_radec=torch.ones(2),
        )
        with self.assertRaises(ap.errors.ConflicingWCS):
            test_image = ap.image.Image_List((base_image1, bad_base_image2))

        # Bad tangent plane x y reference point
        bad_base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            reference_planexy=torch.ones(2),
        )
        with self.assertRaises(ap.errors.ConflicingWCS):
            test_image = ap.image.Image_List((base_image1, bad_base_image2))

        # Bad WCS projection
        bad_base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            projection="orthographic",
        )
        with self.assertRaises(ap.errors.ConflicingWCS):
            test_image = ap.image.Image_List((base_image1, bad_base_image2))


class TestModelImageList(unittest.TestCase):
    def test_model_image_list_creation(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Model_Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Model_Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )

        test_image = ap.image.Model_Image_List((base_image1, base_image2))

        save_image = test_image.copy()
        second_image = test_image.copy()

        second_image += (2, 2)
        second_image -= (1, 1)

        test_image += second_image

        test_image -= second_image

        self.assertTrue(
            torch.all(test_image[0].data == save_image[0].data),
            "adding then subtracting should give the same image",
        )
        self.assertTrue(
            torch.all(test_image[1].data == save_image[1].data),
            "adding then subtracting should give the same image",
        )

        print(test_image.data)
        test_image.clear_image()
        print(test_image.data)
        test_image.replace(second_image)
        print(test_image.data)

        test_image -= (1, 1)
        print(test_image.data)

        self.assertTrue(
            torch.all(test_image[0].data == save_image[0].data),
            "adding then subtracting should give the same image",
        )
        self.assertTrue(
            torch.all(test_image[1].data == save_image[1].data),
            "adding then subtracting should give the same image",
        )

        self.assertIsNone(
            test_image.target_identity,
            "Targets have not been assigned so target identity should be None",
        )

    def test_errors(self):

        # Model_Image_List with non Model_Image object
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Model_Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Target_Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )

        with self.assertRaises(ap.errors.InvalidImage):
            test_image = ap.image.Model_Image_List((base_image1, base_image2))
        

class TestTargetImageList(unittest.TestCase):
    def test_target_image_list_creation(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Target_Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
            variance=torch.ones_like(arr1),
            mask=torch.zeros_like(arr1),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Target_Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
            variance=torch.ones_like(arr2),
            mask=torch.zeros_like(arr2),
        )

        test_image = ap.image.Target_Image_List((base_image1, base_image2))

        save_image = test_image.copy()
        second_image = test_image.copy()

        second_image += (2, 2)
        second_image -= (1, 1)

        test_image += second_image

        test_image -= second_image

        self.assertTrue(
            torch.all(test_image[0].data == save_image[0].data),
            "adding then subtracting should give the same image",
        )
        self.assertTrue(
            torch.all(test_image[1].data == save_image[1].data),
            "adding then subtracting should give the same image",
        )

        test_image += (1, 1)
        test_image -= (1, 1)

        self.assertTrue(
            torch.all(test_image[0].data == save_image[0].data),
            "adding then subtracting should give the same image",
        )
        self.assertTrue(
            torch.all(test_image[1].data == save_image[1].data),
            "adding then subtracting should give the same image",
        )

    def test_targetlist_errors(self):
        arr1 = torch.zeros((10, 15))
        base_image1 = ap.image.Target_Image(
            data=arr1,
            pixelscale=1.0,
            zeropoint=1.0,
            origin=torch.zeros(2),
            variance=torch.ones_like(arr1),
            mask=torch.zeros_like(arr1),
        )
        arr2 = torch.ones((15, 10))
        base_image2 = ap.image.Image(
            data=arr2,
            pixelscale=0.5,
            zeropoint=2.0,
            origin=torch.ones(2),
        )
        with self.assertRaises(ap.errors.InvalidImage):
            test_image = ap.image.Target_Image_List((base_image1, base_image2))
        

class TestJacobianImageList(unittest.TestCase):
    def test_jacobian_image_list_creation(self):
        arr1 = torch.zeros((10, 15, 3))
        base_image1 = ap.image.Jacobian_Image(
            data=arr1,
            parameters=["a", "b", "c"],
            target_identity="target1",
            pixelscale=1.0,
            zeropoint=1.0,
            window=ap.image.Window(
                origin=torch.zeros(2) + 0.1, pixel_shape=torch.tensor((15, 10))
            ),
        )
        arr2 = torch.ones((15, 10, 3))
        base_image2 = ap.image.Jacobian_Image(
            data=arr2,
            parameters=["a", "b", "c"],
            target_identity="target2",
            pixelscale=0.5,
            zeropoint=2.0,
            window=ap.image.Window(
                origin=torch.zeros(2) + 0.2, pixel_shape=torch.tensor((10, 15))
            ),
        )

        test_image = ap.image.Jacobian_Image_List((base_image1, base_image2))

        second_image = test_image.copy()

        test_image += second_image

        self.assertEqual(
            test_image.flatten("data").shape,
            (300, 3),
            "flattened jacobian should include all pixels and merge parameters",
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import astrophot as ap
import torch
import numpy as np
from utils import make_basic_sersic, make_basic_gaussian, make_basic_gaussian_psf


class TestGroup(unittest.TestCase):
    def test_groupmodel_creation(self):
        np.random.seed(12345)
        shape = (10, 15)
        tar = ap.image.Target_Image(
            data=np.random.normal(loc=0, scale=1.4, size=shape),
            pixelscale=0.8,
            variance=np.ones(shape) * (1.4 ** 2),
        )

        mod1 = ap.models.Component_Model(
            name="base model 1",
            target=tar,
            parameters={"center": {"value": [5, 5], "locked": True}},
        )
        mod2 = ap.models.Component_Model(
            name="base model 2",
            target=tar,
            parameters={"center": {"value": [5, 5], "locked": True}},
        )

        smod = ap.models.AstroPhot_Model(
            name="group model",
            model_type="group model",
            models=[mod1, mod2],
            target=tar,
        )

        self.assertFalse(smod.locked, "default model state should not be locked")

        smod.initialize()

        self.assertTrue(torch.all(smod().data == 0), "model_image should be zeros")

    def test_jointmodel_creation(self):
        np.random.seed(12345)
        shape = (10, 15)
        tar = ap.image.Target_Image(
            data=np.random.normal(loc=0, scale=1.4, size=shape),
            pixelscale=0.8,
            variance=np.ones(shape) * (1.4 ** 2),
        )
        shape2 = (33, 42)
        tar2 = ap.image.Target_Image(
            data=np.random.normal(loc=0, scale=1.4, size=shape2),
            pixelscale=0.3,
            origin = (43.2, 78.01),
            variance=np.ones(shape2) * (1.4 ** 2),
        )

        mod1 = ap.models.Flat_Sky(
            name="base model 1",
            target=tar,
        )
        mod2 = ap.models.Flat_Sky(
            name="base model 2",
            target=tar2,
        )

        smod = ap.models.AstroPhot_Model(
            name="group model",
            model_type="group model",
            models=[mod1, mod2],
            target=tar,
        )

        self.assertFalse(smod.locked, "default model state should not be locked")

        smod.initialize()

        self.assertTrue(torch.all(torch.isfinite(smod().data)), "model_image should be real")

    def test_groupmodel_saveload(self):
        np.random.seed(12345)
        tar = make_basic_sersic(N = 51, M=51)

        psf = ap.models.Moffat_PSF(
            name="psf model 1",
            target=make_basic_gaussian_psf(N = 11),
            parameters={
                "center": {"value": [5, 5], "locked": True},
                "n": 2.,
                "Rd": 3.,
                "I0": {"value": 0., "locked": True},
            },
        )
        
        mod1 = ap.models.Sersic_Galaxy(
            name="base model 1",
            target=tar,
            parameters={"center": {"value": [5, 5], "locked": False}},
            psf = psf,
            psf_mode = "full",
        )
        mod2 = ap.models.Sersic_Galaxy(
            name="base model 2",
            target=tar,
            parameters={"center": {"value": [5, 5], "locked": False}},
        )

        smod = ap.models.AstroPhot_Model(
            name="group model",
            model_type="group model",
            models=[mod1, mod2],
            target=tar,
        )

        self.assertFalse(smod.locked, "default model state should not be locked")

        smod.initialize()

        self.assertTrue(torch.all(torch.isfinite(smod().data)), "model_image should be real values")

        smod.save("test_save_group_model.yaml")

        newmod = ap.models.AstroPhot_Model(
            name = "group model",
            filename = "test_save_group_model.yaml",
        )
        self.assertEqual(len(smod.models), len(newmod.models), "Group model should load sub models")

        self.assertEqual(newmod.parameters.size, 16, "Group model size should sum all parameters")

        self.assertTrue(torch.all(newmod.parameters.vector_values() == smod.parameters.vector_values()), "Save/load should extract all parameters")
        

class TestPSFGroup(unittest.TestCase):
    def test_psfgroupmodel_creation(self):
        tar = make_basic_gaussian_psf()

        mod1 = ap.models.AstroPhot_Model(
            name="base model 1",
            model_type="moffat psf model",
            target=tar,
        )
        
        mod2 = ap.models.AstroPhot_Model(
            name="base model 2",
            model_type="moffat psf model",
            target=tar,
        )

        smod = ap.models.AstroPhot_Model(
            name="group model",
            model_type="psf group model",
            models=[mod1, mod2],
            target=tar,
        )

        smod.initialize()

        self.assertTrue(torch.all(smod().data >= 0), "PSF group sample should be greater than or equal to zero")

    def test_psfgroupmodel_saveload(self):
        np.random.seed(12345)
        tar = make_basic_gaussian_psf()

        psf1 = ap.models.Moffat_PSF(
            name="psf model 1",
            target=tar,
            parameters={
                "n": 2.,
                "Rd": 3.,
            },
        )
        
        psf2 = ap.models.Sersic_PSF(
            name="psf model 2",
            target=tar,
            parameters={
                "n": 2.,
                "Re": 3.,
            },
        )

        smod = ap.models.AstroPhot_Model(
            name="group model",
            model_type="psf group model",
            models=[psf1, psf2],
            target=tar,
        )

        smod.initialize()

        self.assertTrue(torch.all(torch.isfinite(smod().data)), "psf_image should be real values")

        smod.save("test_save_psfgroup_model.yaml")

        newmod = ap.models.AstroPhot_Model(
            name = "group model",
            filename = "test_save_psfgroup_model.yaml",
        )
        self.assertEqual(len(smod.models), len(newmod.models), "Group model should load sub models")

        self.assertEqual(newmod.parameters.size, 4, "Group model size should sum all parameters")

        self.assertTrue(torch.all(newmod.parameters.vector_values() == smod.parameters.vector_values()), "Save/load should extract all parameters")
        


from .base import AstroPhotError

__all__ = ("InvalidModel", "InvalidTarget", "UnrecognizedModel")

class InvalidModel(AstroPhotError):
    """
    Catches when a model object is inapropriate for this instance.
    """
    ...

class InvalidTarget(AstroPhotError):
    """
    Catches when a target object is assigned incorrectly.
    """
    ...

class UnrecognizedModel(AstroPhotError):
    """
    Raised when the user tries to invoke a model that does not exist.
    """
    ...

